/*
    $Id: average.h 273 2012-02-22 09:42:17Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _AVERAGE_H
#define _AVERAGE_H

#include <cmath>

#include <boost/numeric/ublas/fwd.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

#include "ublas_math.h"

namespace utils {
namespace algo {
	
namespace detail {
	template<class T> struct mean_op {
		typedef T value_type;
		typedef T result_type;
		
		result_type operator() ( const value_type& x ) const {
			return x;
		}
	};
	
	template<class T> struct variance_op {
		typedef T value_type;
		typedef T result_type;

		variance_op( const value_type& mean ): mean(mean) {}
		
		result_type operator() ( const value_type& x ) const {
			return (mean-x)*(mean-x);
		}
		
		value_type mean;
	};
	
	template<class U> struct boost_variance_op {
		typedef U value_type;
		typedef U result_type;
		
		boost_variance_op( const value_type& mean ): mean(mean) {}
		
		result_type operator() ( const value_type& x ) const {
			using namespace boost::numeric::ublas;
			
			return element_prod(mean-x,mean-x);
		}
		
		value_type mean;
	};
		
	template<class U> struct boost_covariance_op {
		typedef U value_type;
		typedef boost::numeric::ublas::symmetric_matrix< typename U::value_type > result_type;
		
		boost_covariance_op( const value_type& mean ): mean(mean) {}
		
		result_type operator() ( const value_type& x ) const {
			using namespace boost::numeric::ublas;
			
			return outer_prod(mean-x,mean-x);
		}
		
		value_type mean;
	};
	
	
	/* fwd */
	template<class T> struct mean_traits;
	template<class T> struct variance_traits;
	template<class T> struct covariance_traits;
	template<class T> struct error_of_mean_traits;
	
	template<class T> struct mean_traits {
		typedef mean_op<T> op_type;
	};
	template<class T> struct variance_traits {
		typedef variance_op<T> op_type;
	};
	template<class T> struct variance_traits< boost::numeric::ublas::vector<T> > {
		typedef boost_variance_op< boost::numeric::ublas::vector<T> > op_type;
	};
	template<class T> struct variance_traits< boost::numeric::ublas::matrix<T> > {
		typedef boost_variance_op< boost::numeric::ublas::matrix<T> > op_type;
	};
	template<class T> struct variance_traits< boost::numeric::ublas::symmetric_matrix<T> > {
		typedef boost_variance_op< boost::numeric::ublas::symmetric_matrix<T> > op_type;
	};
	template<class T> struct variance_traits< boost::numeric::ublas::banded_matrix<T> > {
		typedef boost_variance_op< boost::numeric::ublas::banded_matrix<T> > op_type;
	};
	template<class T> struct covariance_traits< boost::numeric::ublas::vector<T> > {
		typedef boost_covariance_op< boost::numeric::ublas::vector<T> > op_type;
	};
	template<class T> struct error_of_mean_traits {
		typedef typename variance_traits<T>::op_type op_type;
		static inline T sqrt( const T& x ) {
			return std::sqrt(x);
		}
	};
	template<class T> struct error_of_mean_traits< boost::numeric::ublas::vector<T> > {
		typedef boost::numeric::ublas::vector<T> value_type;
		
		typedef typename variance_traits< value_type >::op_type op_type;
		static inline value_type sqrt( const value_type x ) {
			return utils::ublas::sqrt(x);
		}
	};
	template<class T> struct error_of_mean_traits< boost::numeric::ublas::symmetric_matrix<T> > {
		typedef boost::numeric::ublas::symmetric_matrix<T> value_type;

		typedef typename variance_traits< value_type >::op_type op_type;
		static inline value_type sqrt( const value_type x ) {
			return utils::ublas::sqrt(x);
		}
	};
}

/* average */
template<class Iterator, class Op> typename Op::result_type accumulate( Iterator begin, Iterator end, Op op ) {
	typedef typename Op::result_type result_type;
	
	if( begin == end ) return result_type();
	
	result_type result = op(*begin++);
	
	while( begin != end ) {
		result += op(*begin++);
	}
	
	return result;
}

/* mean */
template<class Iterator> typename detail::mean_traits< typename std::iterator_traits< Iterator >::value_type >::op_type::result_type mean( Iterator begin, Iterator end ) {
	typedef typename std::iterator_traits< Iterator >::value_type value_type;
	typedef typename detail::mean_traits< value_type >::op_type op_type;

	return utils::algo::accumulate( begin, end, op_type() ) / std::distance( begin, end );
}

/* variance */
template<class Iterator> typename detail::variance_traits< typename std::iterator_traits< Iterator >::value_type >::op_type::result_type variance( Iterator begin, Iterator end, const typename std::iterator_traits< Iterator >::value_type& mean ) {
	typedef typename std::iterator_traits< Iterator >::value_type value_type;
	typedef typename detail::variance_traits< value_type >::op_type op_type;

	return utils::algo::accumulate( begin, end, op_type(mean) ) / ( std::distance( begin, end ) - 1 );
}

/* covariance */
template<class Iterator> typename detail::covariance_traits< typename std::iterator_traits< Iterator >::value_type >::op_type::result_type covariance( Iterator begin, Iterator end, const typename std::iterator_traits< Iterator >::value_type& mean ) {
	typedef typename std::iterator_traits< Iterator >::value_type value_type;
	typedef typename detail::covariance_traits< value_type >::op_type op_type;

	return utils::algo::accumulate( begin, end, op_type(mean) ) / ( std::distance( begin, end ) - 1 );
}

/* error of mean */
template<class Iterator> typename detail::error_of_mean_traits< typename std::iterator_traits< Iterator >::value_type >::op_type::result_type error_of_mean( Iterator begin, Iterator end, const typename std::iterator_traits< Iterator >::value_type& mean ) {
	typedef typename std::iterator_traits< Iterator >::value_type value_type;
	typedef typename detail::error_of_mean_traits< value_type >::op_type op_type;

	typename std::iterator_traits< Iterator >::difference_type size = std::distance( begin, end );

	return detail::error_of_mean_traits< value_type >::sqrt(utils::algo::accumulate( begin, end, op_type(mean) ) / (size*(size-1)));
}

}}

#endif // _AVERAGE_H
