/*
    $Id: background_factory.cpp 323 2024-10-30 12:07:25Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdexcept>
#include <boost/format.hpp>

#include "background_factory.h"

namespace atmos {
namespace background {

factory::factory(const select_model_function_type& select_model):
	m_curmodel(),
	m_select_model(select_model) {
}
factory::factory(const factory& factory) {
	m_select_model = factory.m_select_model;
	m_curmodel = std::unique_ptr<model_type>(factory.m_curmodel->clone());
}
void factory::swap(factory& f) throw() {
	f.m_select_model.swap( m_select_model );
	std::swap( f.m_curmodel, m_curmodel );
}
factory& factory::operator=(const factory& f){
	factory tmp( f );
	tmp.swap(*this );
	return *this;
}
factory::~factory() {
}
factory::flux_type factory::background( const time_type& time ) const {
	if( ! m_curmodel.get() )
		throw std::logic_error("There is no any background model");
	return m_curmodel->background( time );
}
factory::model_type* factory::create_new_model( const time_type& time ) const {
	return m_select_model( time );
}
factory::model_type* factory::release_current_model() {
	return m_curmodel.release();
}
void factory::handle_new_model( model_type* model ) {
	m_curmodel = std::unique_ptr<model_type>( model );
}

namespace model {

abstract::abstract( const time_type& time ): m_inittime(time) {
}
abstract::~abstract() {
}

daytime::daytime( const time_type& time ):
	abstract(time),
	m_size(0) {
}
daytime::size_type daytime::size() const {
	return m_size;
}
daytime::flux_type daytime::background( const time_type& time ) const {
	throw std::logic_error("Daytime background observation");
}
void daytime::accumulate( const time_type& time, const flux_type& flux ) {
	m_size++;
}
abstract* daytime::clone() const {
	return new daytime(*this);
}

twilight::twilight( const time_type& time, const sun_altitude_function_type& sun_altitude ):
	abstract(time),
	m_size(0),
	m_background(0),
	m_sun_altitude(sun_altitude) {
}
value_type twilight::extrapolate( const time_type& timef ) const {
	value_type hi = m_sun_altitude( inittime() ) * 180.0 / M_PI; // degrees
	value_type hf = m_sun_altitude( timef ) * 180.0 / M_PI; // degrees
	if( hi < -12.0 ) hi = -12.0;
	if( hf < -12.0 ) hf = -12.0;
	value_type scale = 0.52189 - 0.26242*hi - 0.0235464*hi*hi; // model
	return std::exp( scale * ( hf - hi ) );
}
twilight::size_type twilight::size() const {
	return m_size;
}
twilight::flux_type twilight::background( const time_type& time ) const {
	return m_background * extrapolate( time );
}
void twilight::accumulate( const time_type& time, const flux_type& flux ) {

	if( m_size == 0 ) {
		m_background = flux / extrapolate( time );
		m_size++;
		return;
	}

	m_background = m_background * ( value_type( m_size ) / value_type( m_size + 1 ) ) + flux / extrapolate( time ) * ( value_type(1) / value_type( m_size + 1 ) );
	m_size++;
}
abstract* twilight::clone() const {
	return new twilight(*this);
}

night::night( const time_type& time ):
	abstract(time),
	m_size(0),
	m_background(0) {
}
night::size_type night::size() const {
	return m_size;
}
night::flux_type night::background( const time_type& time ) const {
	return m_background;
}
void night::accumulate( const time_type& time, const flux_type& flux ) {
	if( m_size == 0 ) {
		m_background = flux;
		m_size++;
		return;
	}

	assert( flux.size() == m_background.size() );

	m_background = m_background * ( value_type( m_size ) / value_type( m_size + 1 ) ) + flux * ( value_type(1) / value_type( m_size + 1 ) );
	m_size++;
}
abstract* night::clone() const {
	return new night(*this);
}

}

namespace filter {

nonpoisson::nonpoisson( const value_type& threshold, const value_type& nonpoissonD, const value_type& numexp ):
	m_threshold( threshold ),
	m_nonpoissonD( nonpoissonD ),
	m_numexp( numexp ) {
}
nonpoisson::~nonpoisson() {
}
bool nonpoisson::operator() ( const time_type& time, const flux_type& flux, const vars_type& vars ) const {
	double poisson = ( vars(3,3)*flux(3) );
	double relepsp = 2.0 * ( 1.0 + 1.0/flux(3) ) / ( m_numexp );
	double pmax = m_nonpoissonD * ( 1.0 + m_threshold * std::sqrt(relepsp) );
	if( poisson > pmax ) {
		throw std::logic_error( (boost::format("Background nonpoisson %.4f is over the threshold %.4f") % poisson % pmax).str() );
		return true;
	}
	return false;
}

}
}}
