/*
    $Id: background_factory.h 328 2024-10-30 12:37:15Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _BACKGROUND_FACTORY_H
#define _BACKGROUND_FACTORY_H

#include <memory> // for std::unique_ptr
#include <functional>

#include <boost/date_time/posix_time/posix_time.hpp>

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

namespace atmos {
namespace background {

using namespace boost::numeric::ublas;

typedef boost::posix_time::ptime   time_type;
typedef double value_type;
typedef vector<value_type>         vector_type;
typedef vector_type::size_type     size_type;
typedef symmetric_matrix< value_type > symmetric_matrix_type;

typedef vector_type           flux_type;
typedef symmetric_matrix_type vars_type;

namespace model {
	class abstract;
}

class factory {
public:
	typedef model::abstract              model_type;
	typedef atmos::background::time_type time_type;
	typedef atmos::background::flux_type flux_type;
	
	typedef std::function<model_type* (const time_type&)> select_model_function_type;
private:
	std::unique_ptr< model_type > m_curmodel;
	select_model_function_type  m_select_model;
public:
	factory(const select_model_function_type& select_model);
	factory(const factory& factory);
	void swap(factory& factory) throw();
	factory& operator=(const factory& factory);
	~factory();

	flux_type background( const time_type& time ) const;

	model_type* create_new_model( const time_type& time ) const;
	model_type* release_current_model();
	void handle_new_model( model_type* model );
};

namespace model {
	
class abstract {
public:
	typedef atmos::background::size_type size_type;
	typedef atmos::background::time_type time_type;
	typedef atmos::background::flux_type flux_type;
private:
	time_type m_inittime;
public:
	abstract( const time_type& time );
	virtual ~abstract() = 0;

	const time_type& inittime() const { return m_inittime; }

	virtual size_type size() const = 0;
	virtual flux_type background( const time_type& time ) const = 0;
	virtual void accumulate( const time_type& time, const flux_type& flux ) = 0;

	virtual abstract* clone() const = 0;
};

class daytime: public abstract {
private:
	size_type   m_size;
public:
	daytime( const time_type& time );

	size_type size() const;
	flux_type background( const time_type& time ) const;
	void accumulate( const time_type& time, const flux_type& flux );

	abstract* clone() const;
};

class twilight: public abstract {
public:
	typedef std::function<value_type (const time_type&)> sun_altitude_function_type;
private:
	size_type   m_size;
	flux_type   m_background;
	sun_altitude_function_type m_sun_altitude;
private:
	value_type extrapolate( const time_type& timef ) const;
public:
	twilight( const time_type& time, const sun_altitude_function_type& sun_altitude );

	size_type size() const;
	flux_type background( const time_type& time ) const;
	void accumulate( const time_type& time, const flux_type& flux );

	abstract* clone() const;
};

class night: public abstract {
private:
	size_type   m_size;
	flux_type m_background;
public:
	night( const time_type& time );

	size_type size() const;
	flux_type background( const time_type& time ) const;
	void accumulate( const time_type& time, const flux_type& mean );

	abstract* clone() const;
};
}

namespace filter {

class nonpoisson {
public:
	typedef atmos::background::time_type  time_type;
	typedef atmos::background::flux_type  flux_type;
	typedef atmos::background::vars_type  vars_type;
	typedef atmos::background::value_type value_type;
private:
	value_type m_threshold;
	value_type m_nonpoissonD;
	value_type m_numexp;
public:
	nonpoisson( const value_type& threshold, const value_type& nonpoissonD, const value_type& numexp );
	~nonpoisson();

	bool operator() ( const time_type& time, const flux_type& flux, const vars_type& vars ) const;
};

}
}}

#endif // _BACKGROUND_FACTORY_H
