/*
    $Id: covariance_decomposition.h 329 2024-10-30 12:52:28Z matwey $
    Copyright (C) 2011  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _COVARIANCE_DECOMPOSITION_H
#define _COVARIANCE_DECOMPOSITION_H

#include <cmath>
#include <tuple>

#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/triangular.hpp>

#include <lsp/cholesky_decomposition.h>

#include "ublas_math.h"

namespace utils {
namespace ublas {

using namespace boost::numeric::ublas;

namespace detail {
	template<class M> std::tuple< banded_matrix< typename M::value_type >, symmetric_matrix< typename M::value_type > > correlation_matrix( const M& cov ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		symmetric_matrix< value_type > cor(cov);
		banded_matrix< value_type > sigma = utils::ublas::sqrt( banded_adaptor< symmetric_matrix< value_type > >(cor) );
		for( typename banded_matrix< value_type >::size_type i = 0; i < sigma.size1(); ++i ) {
			row(cor,i)    /= sigma(i,i);
			column(cor,i) /= sigma(i,i);
		}

		return std::make_tuple( sigma, cor );
	}
	template<class M> triangular_matrix< typename M::value_type > sqrt_matrix_decomposition( const M& cov ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		symmetric_matrix< value_type > decomp(cov);
		lsp::cholesky_decomposition( decomp );
		triangular_adaptor< symmetric_matrix< value_type > > result(decomp);
		
		return result;
	}
	template<class M> triangular_matrix< typename M::value_type > inv_sqrt_matrix_decomposition( const M& cov ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		triangular_matrix< value_type > triang( sqrt_matrix_decomposition(cov) );
		matrix< value_type > inv = solve( triang, identity_matrix< value_type >( triang.size1() ), lower_tag() );

		return triangular_adaptor< matrix< value_type > >(inv);
	}
	template<class M> triangular_matrix< typename M::value_type > aetkin_weight_matrix( const M& cov, const typename M::value_type& cond_threshold = 10000 ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		if( BOOST_UBLAS_SAME( cov.size1(), cov.size2() ) < 1 )
			return triangular_matrix< value_type >(0,0);

		for( size_type i = 0; i < BOOST_UBLAS_SAME( cov.size1(), cov.size2() ); ++i ) {
			if( cov(i,i) <= value_type(0) )
				throw std::logic_error("Non-positive diagonal value");
		}

		std::tuple< banded_matrix< value_type >, symmetric_matrix< value_type > > sigma_cor = correlation_matrix(cov);
		triangular_matrix< value_type > weight = sqrt_matrix_decomposition( std::get<1>(sigma_cor) );

		value_type min = std::abs( weight(0,0) );
		value_type max = std::abs( weight(0,0) );
		for( typename triangular_matrix< value_type >::size_type i = 1; i < weight.size1(); ++i ) {
			if( !std::isfinite(weight(i,i)) ) throw std::logic_error( "Infinite diagonal value" );
			if( std::abs( weight(i,i) ) > max ) max = std::abs( weight(i,i) );
			if( std::abs( weight(i,i) ) < min ) min = std::abs( weight(i,i) );
		}

		value_type cond = max / min;
		if( !std::isfinite( cond ) || cond > cond_threshold ) {
			throw std::logic_error( (boost::format("The condition number of covariance matrix is %1%") % cond).str() );
		}

		matrix< value_type > inv = solve( weight, identity_matrix< value_type >( weight.size1() ), lower_tag() );
		inv = prod( inv, utils::ublas::inv( std::get<0>(sigma_cor) ) );
		return triangular_adaptor< matrix< value_type > >(inv);
	}
}

template<class T> std::tuple< banded_matrix<T>, symmetric_matrix<T> > correlation_matrix( const symmetric_matrix<T>& cov ) {
	return detail::correlation_matrix( cov );
}
template<class T> triangular_matrix<T> sqrt_matrix_decomposition( const symmetric_matrix<T>& cov ) {
	return detail::sqrt_matrix_decomposition( cov );
}
template<class T> triangular_matrix<T> inv_sqrt_matrix_decomposition( const symmetric_matrix<T>& cov ) {
	return detail::inv_sqrt_matrix_decomposition( cov );
}
template<class T> triangular_matrix<T> aetkin_weight_matrix( const symmetric_matrix<T>& cov ) {
	return detail::aetkin_weight_matrix( cov );
}

template<class M> std::tuple< banded_matrix< typename M::value_type >, symmetric_matrix< typename M::value_type > > correlation_matrix( const symmetric_adaptor<M>& cov ) {
	return detail::correlation_matrix( cov );
}
template<class M> triangular_matrix< typename M::value_type > sqrt_matrix_decomposition( const symmetric_adaptor<M>& cov ) {
	return detail::sqrt_matrix_decomposition( cov );
}
template<class M> triangular_matrix< typename M::value_type > inv_sqrt_matrix_decomposition( const symmetric_adaptor<M>& cov ) {
	return detail::inv_sqrt_matrix_decomposition( cov );
}
template<class M> triangular_matrix< typename M::value_type >  aetkin_weight_matrix( const symmetric_adaptor<M>& cov ) {
	return detail::aetkin_weight_matrix( cov );
}

}}

#endif // _COVARIANCE_DECOMPOSITION_H
