/*
    $Id: dimm.h 180 2011-01-30 15:56:33Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _DIMM_H
#define _DIMM_H

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "table_storage.h"

namespace atmos {
namespace dimm {

using namespace boost::numeric::ublas;

typedef boost::posix_time::ptime time_type;
typedef double                   value_type;
typedef vector<value_type>       vector_type;
typedef vector_type::size_type   size_type;

typedef vector_type  flux_type;
typedef vector_type  vars_type;
typedef unsigned int franenum_type;

vars_type  correction( const vars_type& vars );
value_type turbulence( value_type scale, value_type Knumber, value_type variance );
value_type seeing( value_type turbulence_power );

template<class Iterator, class T> void correct_low_vars( Iterator begin, Iterator end, const T& mean ) {
	typedef Iterator iterator_type;
	typedef T value_type;

	for( iterator_type it = begin; it != end; ++it ) {
		(*it)(1) += ((*it)(0)-mean)*((*it)(0)-mean);
	}
}

namespace detail {

struct sample {
	typedef atmos::dimm::time_type     timestamp_type;
	typedef atmos::dimm::franenum_type framenum_type;
	typedef atmos::dimm::flux_type     flux_type;
	typedef atmos::dimm::vars_type     vars_type;
		
	timestamp_type timestamp;
	framenum_type  framenum;
	flux_type      left_flux;
	flux_type      right_flux;
	vars_type      xvars;
	vars_type      yvars;

	sample( const timestamp_type& time,
		     const framenum_type& framenum,
		     const flux_type& left,
		     const flux_type& right,
		     const vars_type& xvars,
		     const vars_type& yvars ):
		timestamp(time), framenum(framenum), left_flux(left), right_flux(right), xvars(xvars), yvars(yvars) {}
};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef storage self_type;
	typedef table_storage_column_proxy< self_type, detail::sample::flux_type > left_flux_column_type;
	typedef table_storage_column_proxy< self_type, detail::sample::flux_type > right_flux_column_type;
	typedef table_storage_column_proxy< self_type, detail::sample::vars_type > xvars_column_type;
	typedef table_storage_column_proxy< self_type, detail::sample::vars_type > yvars_column_type;
private:
	left_flux_column_type  m_left_flux_column;
	right_flux_column_type m_right_flux_column;
	xvars_column_type      m_xvars_column;
	yvars_column_type      m_yvars_column;
public:
	storage( size_type reserve = 60 );

	const left_flux_column_type&  left_flux()  const { return m_left_flux_column; }
	const right_flux_column_type& right_flux() const { return m_right_flux_column; }
	const xvars_column_type&      xvars()      const { return m_xvars_column; }
	const yvars_column_type&      yvars()      const { return m_yvars_column; }
	left_flux_column_type&  left_flux()  { return m_left_flux_column; }
	right_flux_column_type& right_flux() { return m_right_flux_column; }
	xvars_column_type&      xvars()      { return m_xvars_column; }
	yvars_column_type&      yvars()      { return m_yvars_column; }

	void push_back( const time_type& time,
			const franenum_type& framenum,
			const flux_type& left,
			const flux_type& right,
			const vars_type& xvars,
			const vars_type& yvars ) {

		table_storage< detail::sample >::push_back( detail::sample( time, framenum, left, right, xvars, yvars ) );
	}
};

}}

#endif // _DIMM_H
