/*
    $Id: fft.h 199 2011-02-03 09:38:16Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _FFT_H
#define _FFT_H

#include <cmath>
#include <complex>

#include <boost/mpl/if.hpp>
#include <boost/numeric/ublas/vector_expression.hpp>
#include <boost/type_traits/is_complex.hpp>

namespace utils {
namespace ublas {

using namespace boost::numeric::ublas;

namespace detail {

	template<class T> T nearest_log2( const T& size ) {
		typedef T size_type;

		size_type n = 0;
		size_type size_ = 1;

		do {
			++n;
			size_ *= 2;
		} while( size_ <= size );

		return n - 1;
	}

	template<class V, class E> void copy_bitreversal( V& v, const vector_expression<E>& e ) {
		typedef typename V::size_type size_type;

		size_type size = BOOST_UBLAS_SAME( v.size(), e().size() );

		if( size < 1 ) return;

		assert( std::pow( size_type(2), nearest_log2( size ) ) == size );

		size_type max_step = size / 2;
		size_type cur_step;

		v(0) = e()(0);
		for( size_type i = 1, j = 0; i < size - 1; ++i ) {
			cur_step = max_step;
			while ( j >= cur_step ) {
				j -= cur_step;
				cur_step /= 2;
			}
			j += cur_step;
			v(i) = e()(j);
		}
		v(size - 1) = e()(size - 1);
	}

	template<class V> void swap_bitreversal( V& v ) {
		typedef typename V::size_type size_type;

		using namespace std;

		size_type size = v.size();

		if( size < 1 ) return;

		assert( std::pow( size_type(2), nearest_log2( size ) ) == size );

		size_type max_step = size / 2;
		size_type cur_step;

		for( size_type i = 1, j = 0; i < size - 1; ++i ) {
			cur_step = max_step;
			while ( j >= cur_step ) {
				j -= cur_step;
				cur_step /= 2;
			}
			j += cur_step;
			if( i < j ) swap( v(i), v(j) );
		}
	}

	template<class E> class fft_radix2_result:
		public vector_expression< fft_radix2_result<E> > {
	public:
		typedef E expression_type;
		typedef typename expression_type::size_type size_type;
		typedef typename boost::mpl::if_< boost::is_complex< typename expression_type::value_type >,
			typename expression_type::value_type,
			std::complex< typename expression_type::value_type > >::type complex_type;
		typedef typename complex_type::value_type real_type;
		typedef complex_type value_type;
	private:
		template<class V> void copy_helper( V& v, const vector_expression<E>& e ) const {
			copy_bitreversal( v, e );
		}
		void copy_helper( E& v, const vector_expression<E>& e ) const {
			if( &v == &e() ) {
				swap_bitreversal( v );
				return;
			}
			copy_bitreversal( v, e );
		}
	public:
		fft_radix2_result( const vector_expression<E>& e, bool sign = true ): e_(e), sign_(sign) {
		}
		size_type size() const { return e_().size(); }

		template<class V> void apply( V& v ) const {
			typedef typename V::size_type size_type;
			using namespace std;

			copy_helper( v, e_ );

			size_type size = v.size();
			size_type n = nearest_log2( size );

			for( size_type i = 0, step = 1; i < n; ++i, step+=step ) {
				real_type phase_step = ( sign_ ? M_PI : -M_PI ) / step;
				real_type phase( 0 );

				for ( size_type j = 0; j < step; j++ ) {
					complex_type e( exp( complex_type( 0, phase ) ) );
					phase += phase_step;

					for ( size_type k = j; k < size; k += step*2 ) {
						complex_type t( e * v( k + step ) );
						v( k + step ) = v(k) - t;
						v(k) = v(k) + t;
					}
				}
			}
		}
	private:
		const vector_expression<E>& e_;
		bool sign_;
	};

	template<template <class T1, class T2> class F, class V, class E >
	void vector_assign (V &v, const vector_expression< fft_radix2_result< E > >& e ) {
		e().apply( v );
	}

}

template<class E> detail::fft_radix2_result<E> fft_radix2( const vector_expression<E>& e, bool sign = true ) {
	return detail::fft_radix2_result<E>(e,sign);
}

}
}

#endif // _FFT_H
