/*
    $Id: file_output.cpp 288 2014-10-02 07:39:38Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/format.hpp>
#include <boost/io/ios_state.hpp>

#include "atmos.h"
#include "file_output.h"
#include "io.h"

namespace atmos {
namespace output {

file::file( std::ostream& stm ): m_stm(stm) {
	boost::posix_time::time_facet* facet(new boost::posix_time::time_facet("%Y-%m-%d %H:%M:%S"));
	stm.imbue(std::locale(stm.getloc(), facet));
}
file::~file() {

}
void file::comment( const std::string& comment ) {
	m_stm << "# " << comment << std::endl;
}
void file::normal( const time_type& time, double mz ) {
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "M " << time << " Normal " << std::fixed << std::setprecision(3) << mz << std::endl;
}
void file::background( const time_type& time ) {
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "M " << time << " Background measurements" << std::endl;
}
void file::atmosphere( const time_type& time, double free_seeing, double free_seeing_rerr, double dimm_seeing, double dimm_seeing_rerr, double dimm_turbulence, double dimm_turbulence_rerr, double isoplanatic_angle, double isoplanatic_angle_rerr, double altitude_eff, double altitude_eff_rerr, double desi_time_constant, double desi_time_constant_rerr, double time_constant, double time_constant_rerr, double moment_0, double moment_0_rerr, double moment_2, double moment_2_rerr ) {
	static datum skipped;
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "A " << time << " "
		<< std::fixed << datum( free_seeing, free_seeing_rerr ) << " " << datum( dimm_seeing, dimm_seeing_rerr ) << " "
		<< std::scientific << datum( moment_0, moment_0_rerr ) << " " << datum( dimm_turbulence, dimm_turbulence_rerr ) << " "
		<< std::fixed << datum( altitude_eff, altitude_eff_rerr ) << " " << skipped << " "
		<< datum( isoplanatic_angle, isoplanatic_angle_rerr ) << " "
		<< std::scientific << datum( moment_2, moment_2_rerr ) << " "
		<< std::fixed << datum( desi_time_constant, desi_time_constant_rerr ) << " "
		<< std::fixed << datum( time_constant, time_constant_rerr ) << std::endl;
}
void file::profile( const time_type& time, const std::string& method, double cn2seeing, double chi2, const cn2_type& cn2, const cn2_type& cn2_err, const grid_type& grid ){
	assert( cn2.size() == ecn2.size() );
	assert( cn2.size() == grid.size() );

	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "T " << time << " " << method << " "
		<< std::setw(2) << cn2.size() << " " << std::setprecision(2)
		<< std::fixed << chi2 << " " << cn2seeing;
	for( cn2_type::size_type i = 0; i < BOOST_UBLAS_SAME(cn2.size(),cn2_err.size()); ++i ) {
		m_stm << " " << std::fixed << grid(i) << std::scientific  << " " << cn2(i) << " " << cn2_err(i);
	}
	m_stm << std::endl;
}
void file::residual( const time_type& time, const std::string& method, const residual_type& residual, const boost::optional<std::pair<double, double> >& residual_dimm ) {
	assert( residual.size() == atmos::indices_size  );

	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "R " << time << " " << method << std::setprecision(3) << std::fixed;
	for( residual_type::size_type i = 0; i < BOOST_UBLAS_SAME(residual.size1(),residual.size2()); ++i ) {
		m_stm << " " << residual(i,i);
	}
	for( residual_type::size_type i = 0; i < residual.size1() - 1; ++i ) {
		for( residual_type::size_type j = i + 1; j < residual.size2(); ++j ) {
			m_stm << " " << residual(i,j);
		}
	}
	if( residual_dimm ) {
		m_stm << " " << residual_dimm->first << " " << residual_dimm->second;
	}
	m_stm << std::endl;
}
void file::fluxes( const time_type& time, const flux_type& result_flux ) {
	assert( result_flux.size() == atmos::channels  );

	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "F " << time << std::fixed;
	for( flux_type::size_type i = 0; i < result_flux.size(); ++i ) {
		m_stm << " " << datum(result_flux(i));
	}
	m_stm << std::endl;
}
void file::fluxes( const time_type& time, const vector_type& result_flux, const flux_type& result_flux_err ) {
	assert( result_flux.size() == atmos::channels );
	assert( result_flux.size() == result_flux_err.size() );

	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "F " << time << std::fixed;
	for( flux_type::size_type i = 0; i < BOOST_UBLAS_SAME(result_flux.size(),result_flux_err.size()); ++i ) {
		m_stm << " " << datum(result_flux(i),result_flux_err(i)/result_flux(i));
	}
	m_stm << std::endl;
}
void file::indices( const time_type& time, const indices_type& result_indx, const indices_type& result_indx_err, const indesis_type& result_desi, const indesis_type& result_desi_err ) {
	assert( result_desi.size() == atmos::channels );
	assert( result_indx.size1() == atmos::channels );
	assert( result_indx.size2() == atmos::channels );
	assert( result_desi.size() == result_desi_err.size() );
	assert( result_indx.size1() == result_indx_err.size1() );
	assert( result_indx.size1() == result_indx_err.size2() );

	boost::io::ios_all_saver asaver( m_stm );
	ios_indic_saver isaver( m_stm );

	m_stm << "I " << time << std::indic << std::fixed;
	for( indices_type::size_type i = 0; i < BOOST_UBLAS_SAME(result_indx.size1(),result_indx.size2()); ++i ) {
		m_stm << " " << datum( result_indx(i,i), result_indx_err(i,i)/result_indx(i,i) );
	}
	for( indices_type::size_type i = 0; i < BOOST_UBLAS_SAME(result_indx.size1(),result_indx_err.size1()) - 1; ++i ) {
		for( indices_type::size_type j = i + 1; j < BOOST_UBLAS_SAME(result_indx.size2(),result_indx_err.size2()); ++j ) {
			m_stm << " " << datum( result_indx(i,j), result_indx_err(i,j)/result_indx(i,j) );
		}
	}
	for( indesis_type::size_type i = 0; i < BOOST_UBLAS_SAME(result_desi.size(),result_desi_err.size()); ++i ) {
		m_stm << " " << datum( result_desi(i), result_desi_err(i)/result_desi(i) );
	}
	m_stm << std::endl;
}
void file::parameter( const time_type& time, const std::string& key, const std::string& value ) {
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "P " << time << " " << key << " = " << value << std::endl;
}
void file::objectinfo( const time_type& time, const object_type& obj ){
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "O " << time << " " << obj << std::endl;
}

}}
