/*
    $Id: indices.h 329 2024-10-30 12:52:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _INDICES_H
#define _INDICES_H

#include <tuple>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "table_storage.h"

namespace atmos {
namespace indices {
	using namespace boost::numeric::ublas;

	typedef boost::posix_time::ptime time_type;
	typedef double                   value_type;
	typedef vector< value_type >     vector_type;
	typedef vector_type::size_type   size_type;
	typedef symmetric_matrix< value_type > symmetric_matrix_type;
	
	typedef vector_type           flux_type;
	typedef symmetric_matrix_type vars_type;
	typedef symmetric_matrix_type rho1_type;
	typedef vector_type           rho2_type;
	typedef vector_type           indesis_type;
	typedef symmetric_matrix_type indices_type;

	flux_type flux_nonlinearity_correction( const flux_type& flux, const vector_type& deadtime );
	flux_type flux_background_correction( const flux_type& flux, const vector_type& scatter, const vector_type& background );
	std::tuple< vars_type, rho1_type, rho2_type > vars_nonlinearity_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2, const vector_type& deadtime, const vector_type& npoisson, const value_type& exposition );
	vars_type vars_nonlinearity_correction( const flux_type& flux, const vars_type& vars, const vector_type& deadtime, const vector_type& npoisson, const value_type& exposition );
	vars_type vars_exposure_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1 );
	indesis_type calculate_indesis( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 );
	indices_type calculate_indices( const flux_type& flux, const vars_type& vars );

namespace detail {
	struct sample {
		typedef atmos::indices::time_type    timestamp_type;
		typedef atmos::indices::flux_type    flux_type;
		typedef atmos::indices::indices_type indices_type;
		typedef atmos::indices::indesis_type indesis_type;

		timestamp_type timestamp;
		flux_type      flux;
		indices_type   indices;
		indesis_type   indesis;
		indices_type   wind_indices;

		sample( const timestamp_type& time,
			const flux_type& flux,
			const indices_type& indices,
			const indesis_type& indesis,
			const indices_type& wind_indices ):
			timestamp(time), flux(flux), indices(indices), indesis(indesis), wind_indices(wind_indices) {}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef storage self_type;
	typedef table_storage_column_getter< self_type, detail::sample::flux_type >    flux_column_type;
	typedef table_storage_column_getter< self_type, detail::sample::indices_type > indices_column_type;
	typedef table_storage_column_getter< self_type, detail::sample::indesis_type > indesis_column_type;

private:
	flux_column_type    m_flux_column;
	indices_column_type m_indices_column;
	indesis_column_type m_indesis_column;
	indices_column_type m_wind_indices_column;

public:
	storage( size_type reserve = 60 );

	const flux_column_type&    flux()    const { return m_flux_column; }
	const indices_column_type& indices() const { return m_indices_column; }
	const indesis_column_type& indesis() const { return m_indesis_column; }
	const indices_column_type& wind_indesis() const { return m_wind_indices_column; }

	void push_back( const time_type& time,
			const flux_type& flux,
			const indices_type& indices,
			const indesis_type& indesis,
			const indices_type& wind_indices ) {
		
		table_storage< detail::sample >::push_back( detail::sample( time, flux, indices, indesis, wind_indices ) );
	}
};

namespace filter {
	class flux_threshold {
	public:
		typedef atmos::indices::flux_type  flux_type;
		typedef atmos::indices::value_type value_type;
	private:
		value_type m_threshold;
	public:
		flux_threshold( value_type threshold );
		~flux_threshold();

		bool operator() ( const flux_type& flux ) const;
	};
}
}}

#endif // _INDICES_H
