/*
    $Id: integrals.cpp 332 2024-11-02 07:42:55Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <functional>

#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

#include <lsp/singular_decomposition.h>

#include "atmos.h"
#include "ublas_math.h"
#include "integrals.h"
#include "vector_generator.h"
#include "utils.h"
#include "weif.h"

namespace atmos {
namespace integrals {

value_type seeing( const value_type& moment0 ) {
	const value_type power = 3.0/5.0;
	const value_type k_lambda = 1.1486983549970351 * 15.848931924611138; // lambda^-1/5
	const value_type k = 5.307 * 206264.8;

	return k*k_lambda*std::pow( moment0, power );
}
value_type effective_altitude( const value_type& moment0, const value_type& moment1 ) {
	const value_type k = 0.001; // km in meter

	return k*moment1/moment0;
}
value_type isoplanatic_angle( const value_type& moment53 ) {
	const value_type power = -3.0/5.0;
	const value_type k_lambda = 0.43527528164806206 * 6.3095734448019363e-08; // lambda^6/5
	const value_type k = 0.058 * 206264.8;

	return k*k_lambda*std::pow( moment53, power );
}
value_type time_constant_desi( const value_type& desi, const value_type& exposure ) {
	const value_type power = -3.0/5.0;
	const value_type exposure_power = 5.0/6.0;
	const value_type k = 0.175;

	return k*std::pow( exposure, exposure_power )*std::pow( desi, power );
}
value_type time_constant( const value_type& wind_moment0, const value_type& moment0, const value_type& exposure ) {
	const value_type power = -0.1;
	const value_type k_lambda = 0.43527528164806206 * 6.3095734448019363e-08; // lambda^6/5
	const value_type k = 0.058 * 1e3;
	const value_type k_wind = 6 / ( 6e-6 * exposure*exposure ); // 6 / (2*(2**2-1**2) tau**2 )

	return k*k_lambda*std::pow( moment0, power )/std::sqrt( k_wind*wind_moment0 );
}

std::pair< value_type, value_type > seeing( const value_type& moment0, const value_type& moment0_rerr ) {
	return std::make_pair( seeing(moment0), 3.0/5.0*moment0_rerr );
}
std::pair< value_type, value_type > effective_altitude( const value_type& moment0, const value_type& moment0_rerr, const value_type& moment1, const value_type& moment1_rerr ) {
	return std::make_pair( effective_altitude(moment0,moment1), std::sqrt(moment0_rerr*moment0_rerr+moment1_rerr*moment1_rerr) );
}
std::pair< value_type, value_type > isoplanatic_angle( const value_type& moment53, const value_type& moment53_rerr ) {
	return std::make_pair( isoplanatic_angle(moment53), 3.0/5.0*moment53_rerr );
}
std::pair< value_type, value_type > time_constant_desi( const value_type& desi, const value_type& desi_rerr, const value_type& exposure ) {
	return std::make_pair( time_constant_desi( desi, exposure ), 3.0/5.0*desi_rerr );
}
std::pair< value_type, value_type > time_constant( const value_type& wind_moment0, const value_type& wind_moment0_rerr, const value_type& moment0, const value_type& moment0_rerr, const value_type& exposure ) {
	return std::make_pair( time_constant( wind_moment0, moment0, exposure ), std::sqrt(0.1*moment0_rerr*moment0_rerr + 0.5*wind_moment0_rerr*wind_moment0_rerr) );
}

namespace {

vector_type init_moment_powers() {
	vector_type powers(4);
	powers(moment_0)=0.0;
	powers(moment_53)=5.0/3.0;
	powers(moment_1)=1.0;
	powers(moment_2)=2.0;
	return powers;
}

vector_type init_wind_moment_powers() {
	vector_type powers(1);
	powers(moment_0)=0.0;
	return powers;
}

}

namespace detail {
functor_creator::result_type functor_creator::operator() ( const std::string& spectral_type, const weight_type& null_wf, const weight_type& second_wf ) const {
	static const vector_type moment_powers = init_moment_powers();
	static const vector_type wind_moment_powers = init_wind_moment_powers();

	return result_type( calculate( null_wf, moment_powers ), moment_powers, calculate( second_wf, wind_moment_powers ), wind_moment_powers );
}
matrix_type functor_creator::calculate( const weight_type& wf, const vector_type& powers ) {
	using namespace boost::numeric::ublas;
	using namespace utils::algo;
	using namespace utils::ublas;
	using namespace std::placeholders;

	typedef matrix_type            result_type;
	typedef matrix_type::size_type size_type;
	
	const double max_cond = 1667;
	const double bound_altitude = 0.38; // Height of the boundary layer [kilometers] to fit the Cn2-integral
	const double top_altitude   = 25.0; // Height of the hight layer [kilometers]

	const vector<double> grid( utils::ublas::vector_generator< lg_generator > ( lg_generator(0.0, 0.04, 0.25), 50 ) );
	vector<double>::const_iterator lower = std::find_if( grid.begin(), grid.end(), std::bind(std::greater<double>(), _1, bound_altitude) );
	vector<double>::const_reverse_iterator higher = std::find_if( grid.rbegin(), grid.rend(), std::bind(std::less<double>(), _1, top_altitude) );
	assert( lower != grid.end() ); // both layer altitudes and grid are compile-time contants, so it is assertion.
	assert( higher != grid.rend() );
	vector_range< const vector<double> > h( grid, range( lower.index(), higher.index()+1 ) );
	
	matrix< double > wfv = vector_of_vector_to_matrix( project( h, wf) );

	result_type res( powers.size(), atmos::indices_size );
	
	for( vector_type::size_type ipow = 0; ipow < powers.size(); ipow++ ) {
		double moment_power = powers(ipow);
		double weight_power = moment_power / 2.0;
		double scale = pow( 1000.0, moment_power );

		vector<double> alt    = utils::ublas::pow(h, moment_power) * scale;
		vector<double> weight = utils::ublas::pow(h, weight_power);

		matrix<double> A( trans( subrange( wfv, 0, atmos::indices_size, 0, wfv.size2() ) ) );
		matrix<double> left( identity_matrix<double>( A.size1() ) );
		matrix<double> right( identity_matrix<double>( A.size2() ) );

		for( matrix<double>::size_type i = 0; i < A.size1(); ++i ) {
			alt( i ) /= weight( i );
			row( A, i ) /= weight( i );
		}

		lsp::singular_decomposition< matrix< double > > sd( A );
		sd.apply( left, right );

		matrix_vector_slice< matrix<double> > singular( diag( A ) ); // it is guaranteed that singular values are sorted
		matrix_vector_slice< matrix<double> >::const_iterator small_singular = std::find_if( singular.begin(), singular.end(), std::bind(std::less<double>(), _1, singular(0)/max_cond) );
		matrix_vector_slice< matrix<double> >::const_iterator it = singular.begin();
		alt = prod( left, alt );
		for( ; it != small_singular; ++it ) {
			row( res, ipow )( it.index() ) = alt( it.index() ) / (*it);
		}
		for( ; it != singular.end(); ++it ) {
			row( res, ipow )( it.index() ) = 0;
		}
		row( res, ipow ) = prod( right, row( res, ipow ) );
	}

	return res;
}
functor_key::functor_key( const std::string& spectral_type, const weight_type& null_wf, const weight_type& second_wf ):
	spectral_type(spectral_type) {
}
bool functor_key::operator==( const functor_key& key ) const {
	return !(*this < key) && !(key < *this);
}
bool functor_key::operator!=( const functor_key& key ) const {
	return *this < key || key < *this;
}
bool functor_key::operator< ( const functor_key& key ) const {
	if( nocase_less_( spectral_type, key.spectral_type ) ) {
		return true;
	}
	return false;
}
}

functor::result_type functor::operator() ( const indices_type& indices, const indices_type& wind_indices, const indesis_type& desi, double mz ) const {
	typedef vector< double >       vector_type;
	typedef vector_type::size_type size_type;
	
	assert( indices.size1() == atmos::channels );
	assert( indices.size2() == atmos::channels );
	assert( desi.size()  == atmos::channels );

	vector_type aints = prod( m_indices_dcf, atmos::weight::as_vector(indices) );

	assert( m_powers.size() == aints.size() );
	for( size_type i = 0; i < m_powers.size(); ++i ){
		aints(i) = aints(i) / std::pow( mz, 1.0 + m_powers(i) );
		if( !std::isfinite(aints(i)) || aints(i) < 0.0 )
			aints(i) = 0.0;
	}

	vector_type wind_aints = prod( m_wind_indices_dcf, atmos::weight::as_vector(wind_indices) );
		
	assert( m_wind_powers.size() == wind_aints.size() );
	for( size_type i = 0; i < m_wind_powers.size(); ++i ){
		wind_aints(i) = wind_aints(i) / std::pow( mz, 1.0 + m_wind_powers(i) );
		if( !std::isfinite(wind_aints(i)) || wind_aints(i) < 0.0 )
			wind_aints(i) = 0.0;
	}

	assert( desi.size() > 0 );

	return std::make_tuple( aints, wind_aints, ( !std::isfinite(desi(0)) || desi(0) < 0 ) ? 0 : desi(0) );
}

storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_moments_column( *this, std::mem_fn(&detail::sample::moments) ),
	m_wind_moments_column( *this, std::mem_fn(&detail::sample::wind_moments ) ),
	m_desi_column( *this, std::mem_fn(&detail::sample::desi) ) {
}

}}
