/*
    $Id: integrals.h 329 2024-10-30 12:52:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _INTEGRALS_H
#define _INTEGRALS_H

#include <functional>
#include <tuple>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "caching_factory.h"
#include "indices.h"
#include "nocase_less.h"
#include "table_storage.h"

namespace atmos {
namespace integrals {
	using namespace boost::numeric::ublas;
	
	typedef double value_type;
	typedef vector< value_type > vector_type;
	typedef matrix< value_type > matrix_type;

	typedef vector_type moments_type;
	typedef atmos::indices::indesis_type::value_type desi_type;
	
	typedef std::function< vector<double> (double) > weight_type;

	enum moment_power {
		moment_0 = 0,
		moment_53 = 1,
		moment_1 = 2,
		moment_2 = 3
	};

	value_type seeing( const value_type& moment0 );
	value_type effective_altitude( const value_type& moment0, const value_type& moment1 );
	value_type isoplanatic_angle( const value_type& moment53 );
	value_type time_constant_desi( const value_type& desi, const value_type& exposure );
	value_type time_constant( const value_type& wind_moment0, const value_type& moment0, const value_type& exposure );

	std::pair< value_type, value_type > seeing( const value_type& moment0, const value_type& moment0_rerr );
	std::pair< value_type, value_type > effective_altitude( const value_type& moment0, const value_type& moment0_rerr, const value_type& moment1, const value_type& moment1_rerr );
	std::pair< value_type, value_type > isoplanatic_angle( const value_type& moment53, const value_type& moment53_rerr );
	std::pair< value_type, value_type > time_constant_desi( const value_type& desi, const value_type& desi_rerr, const value_type& exposure );
	std::pair< value_type, value_type > time_constant( const value_type& wind_moment0, const value_type& wind_moment0_rerr, const value_type& moment0, const value_type& moment0_rerr, const value_type& exposure );

namespace detail {
	struct sample {
		typedef atmos::integrals::moments_type moments_type;
		typedef atmos::integrals::desi_type desi_type;

		moments_type moments;
		moments_type wind_moments;
		desi_type desi;

		sample( const moments_type& moments,
			const moments_type& wind_moments,
			const desi_type& desi ):
			moments(moments), wind_moments(wind_moments), desi(desi) {
		}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef storage self_type;
public:
	typedef table_storage_column_proxy< self_type, detail::sample::moments_type > moments_column_type;
	typedef table_storage_column_proxy< self_type, detail::sample::moments_type > wind_moments_column_type;
	typedef table_storage_column_proxy< self_type, detail::sample::desi_type >    desi_column_type;
private:
	moments_column_type m_moments_column;
	wind_moments_column_type m_wind_moments_column;
	desi_column_type m_desi_column;
public:
	storage( size_type reserve = 60 );

	const moments_column_type& moments() const { return m_moments_column; }
	const wind_moments_column_type& wind_moments() const { return m_wind_moments_column; }
	const desi_column_type& desi() const { return m_desi_column; }

	void push_back( const moments_type& moments, const moments_type& wind_moments, const desi_type& desi ) {
		table_storage< detail::sample >::push_back( detail::sample( moments, wind_moments, desi ) );
	}
};

//class functor_factory;
class functor;

namespace detail {
	using namespace utils::algo;
	
	struct functor_creator {
		typedef functor result_type;
		
		result_type operator() ( const std::string& spectral_type, const weight_type& null_wf, const weight_type& second_wf ) const;
		
	private:
		static matrix_type calculate( const weight_type& wf, const vector_type& powers );
	};
	
	struct functor_key {
		functor_key( const std::string& spectral_type, const weight_type& null_wf, const weight_type& second_wf );
	
		bool operator==( const functor_key& key ) const;
		bool operator!=( const functor_key& key ) const;
		bool operator< ( const functor_key& key ) const;
	private:
		const std::string spectral_type;
		
		nocase_less< std::string > nocase_less_;
	};
};

typedef utils::algo::caching_factory< functor, detail::functor_key, detail::functor_creator > functor_factory;

class functor {
	public:
		typedef matrix_type dcf_matrix_type;

		typedef atmos::indices::indices_type indices_type;
		typedef atmos::indices::indesis_type indesis_type;
		typedef std::tuple< atmos::integrals::moments_type, atmos::integrals::moments_type, atmos::integrals::desi_type > result_type;
	private:
		dcf_matrix_type m_indices_dcf;
		const vector_type& m_powers;

		dcf_matrix_type m_wind_indices_dcf;
		const vector_type& m_wind_powers;
	public:
		functor( const dcf_matrix_type& indices_dcf, const vector_type& powers, const dcf_matrix_type& wind_indices_dcf, const vector_type& wind_powers ):
			m_indices_dcf( indices_dcf ), m_powers(powers), m_wind_indices_dcf( wind_indices_dcf ), m_wind_powers(wind_powers) {}

		result_type operator() ( const indices_type& indices, const indices_type& wind_indices, const indesis_type& desi, double mz ) const;
};

}}

#endif // _INTEGRALS_H
