/*
    $Id: lambert.h 329 2024-10-30 12:52:28Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _LAMBERT_H
#define _LAMBERT_H

#include <cmath>
#include <limits>
#include <tuple>

#include <boost/cstdint.hpp>

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
#define HAVE_BOOST_MATH_TOOLS_ROOTS_HPP
#endif

#ifdef HAVE_BOOST_MATH_TOOLS_ROOTS_HPP
#include <boost/math/tools/roots.hpp>
#endif

#ifndef HAVE_BOOST_MATH_TOOLS_ROOTS_HPP
#warning "Please, consider using Boost >= 1.35.0"
#endif


namespace utils {
namespace math {

namespace detail {
	
#ifndef HAVE_BOOST_MATH_TOOLS_ROOTS_HPP
/* An simple implementation of Halley iterator for boost < 1.35.0 */
template <class T,class F> T halley_iterate( F functor, T guess, T, T, int digits, boost::uintmax_t& max_iter ) {
	using namespace std;
	
	T res = guess;
	T f0, f1, f2;
	T dx;
	boost::uintmax_t count(max_iter);
	const T factor( ldexp(1.0, 1 - digits) );
	
	do {
		tr1::tie(f0, f1, f2) = functor( res );
		dx = f0 / ( f1 - f0 * ( f2 / ( 2 * f1 ) ) );
		if( abs(dx/res) < 4 * numeric_limits<T>::epsilon() ) dx = 0;
		res -= dx;
	} while( --count && abs( res * factor ) < abs( dx ) );

	max_iter -= count;
	
	return res;
}
#endif

template <class T>
struct lambert_W0_functor {
	typedef T value_type;

	lambert_W0_functor( T const& target ) : a(target) {
	}
	std::tuple<T, T, T> operator()(T const& z) {
		using namespace std;
		return std::make_tuple( z - a * exp( -z ), z + value_type(1), z + value_type(2) );
	}
	private:
	T a;
};

}

template <class T> T lambert_W0( T z, boost::uintmax_t& max_iter ) {
	using namespace std;
	const T inve = 1.0/M_E;
	T min = -std::numeric_limits<T>::max();
	T max = std::numeric_limits<T>::max();
	T guess = 0.0;
	int digits = std::numeric_limits<T>::digits;
	if( z > M_E ) {
		min   = 1.0;
		guess = log(z)-log( log(z) );
	} else if ( z > inve ) {
		min   = 0;
		guess = 1.0;
	} else if ( z > -0.3 ) {
		guess = z-z*z;
		if( z > 0 )
			min = 0;
		else
			max = 0;
	} else if ( z > -inve ) {
		max   = 0;
		guess = sqrt( 2.0*(M_E*z+1.0) )-1.0;
	} else if ( z == -inve ) {
		return -1.0;
	} else {
		return std::numeric_limits<T>::quiet_NaN();
	}
#ifdef HAVE_BOOST_MATH_TOOLS_ROOTS_HPP
	return boost::math::tools::halley_iterate( detail::lambert_W0_functor<T>(z), guess, min, max, digits, max_iter );
#else
	return detail::halley_iterate( detail::lambert_W0_functor<T>(z), guess, min, max, digits, max_iter );
#endif
}

template <class T> T lambert_W0( T z ) {
	boost::uintmax_t m = (std::numeric_limits<boost::uintmax_t>::max)();
	return lambert_W0( z, m );
}

} // math
} // utils

#ifndef HAVE_CONFIG_H
#undef HAVE_BOOST_MATH_TOOLS_ROOTS_HPP
#endif

#endif // _LAMBERT_H
