/*
    $Id: params.cpp 298 2017-01-08 14:39:07Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdexcept>

#include <sstream>

#include <yalpa/spheric.h>
#include <yalpa/io.h>

#include "atmos.h"
#include "params.h"

namespace atmos {
namespace parameters {

container::container() {}
container::~container() {}
bool container::key_equal( const key_type& k1, const key_type& k2 ) {
	return !m_storage.key_comp()( k1, k2 ) && !m_storage.key_comp()( k2, k1 );
}
bool container::value_equal( const value_type& v1, const value_type& v2 ){
	return !m_storage.value_comp()( v1, v2 ) && !m_storage.value_comp()( v2, v1 );
}
bool container::has(const storage_type::key_type& key) const {
	storage_type::const_iterator it = m_storage.find( key );
	return ( it != m_storage.end() );
}
const container::storage_type::mapped_type& container::at(const storage_type::key_type& key) const {
	storage_type::const_iterator it = m_storage.find( key );
	if( it != m_storage.end() )
		return it->second;
	throw std::range_error( std::string("Not found value for key \"") + key + std::string("\""));
}

proxy::proxy( const container& container ): m_container( container ) {}
proxy::~proxy() {}

value_type proxy::longitude() const {
	value_type longitude = 0.0;
	std::istringstream iss( m_container.at("general/site/longitude") );
	iss.exceptions(std::ios_base::badbit | std::ios_base::failbit);
	iss >> std::sexagesimal >> std::hours >> yalpa::make_angle( longitude );
	return longitude;
}
value_type proxy::latitude() const {
	value_type latitude = 0.0;
	std::istringstream iss( m_container.at("general/site/latitude") );
	iss.exceptions(std::ios_base::badbit | std::ios_base::failbit);
	iss >> std::sexagesimal >> std::degrees >> yalpa::make_angle( latitude );
	return latitude;
}
yalpa::geographic< value_type > proxy::site() const {
	return yalpa::geographic< value_type >( longitude(), latitude() );
}

vector_type proxy::inner_diam()    const {
	typedef vector_type::value_type value_type;
	vector_type ret(atmos::channels);
	ret(0) = m_container.as<value_type>("optics/segment a/inner");
	ret(1) = m_container.as<value_type>("optics/segment b/inner");
	ret(2) = m_container.as<value_type>("optics/segment c/inner");
	ret(3) = m_container.as<value_type>("optics/segment d/inner");
	return ret * ( magnification()*0.1 );
}
vector_type proxy::outer_diam()    const {
	typedef vector_type::value_type value_type;
	vector_type ret(atmos::channels);
	ret(0) = m_container.as<value_type>("optics/segment a/outer");
	ret(1) = m_container.as<value_type>("optics/segment b/outer");
	ret(2) = m_container.as<value_type>("optics/segment c/outer");
	ret(3) = m_container.as<value_type>("optics/segment d/outer");
	return ret * ( magnification()*0.1 );
}
vector_type proxy::nonlinearity() const {
	typedef vector_type::value_type value_type;
	std::vector< std::string > ctr( counters() );
	vector_type ret( ctr.size() );
	for( std::vector< std::string >::size_type i = 0; i < ctr.size(); ++i ) {
		ret(i) = m_container.as< value_type >( (boost::format("Devices/Bicounter %1%/NonLinearity%2%") % ctr[i][0] % ctr[i][1]).str() );
	}
	return ret * 1e-6; // nsec to msec
}
vector_type proxy::nonpoisson() const {
	typedef vector_type::value_type value_type;
	std::vector< std::string > ctr( counters() );
	vector_type ret( ctr.size() );
	for( std::vector< std::string >::size_type i = 0; i < ctr.size(); ++i ) {
		ret(i) = m_container.as< value_type >( (boost::format("Devices/Bicounter %1%/NonPoisson%2%") % ctr[i][0] % ctr[i][1]).str() );
	}
	return ret;
}
vector_type proxy::scatter() const {
	typedef vector_type::value_type value_type;
	vector_type ret(atmos::channels);
	ret(0) = m_container.as< value_type >("channels/scatter/channela",0.0);
	ret(1) = m_container.as< value_type >("channels/scatter/channelb",0.0);
	ret(2) = m_container.as< value_type >("channels/scatter/channelc",0.0);
	ret(3) = m_container.as< value_type >("channels/scatter/channeld",0.0);
	return ret;
}
std::vector< std::string > proxy::counters() const {
	std::vector< std::string > ret;
	ret.resize(atmos::channels);
	ret[0] = m_container.at("channels/configuration/channela");
	ret[1] = m_container.at("channels/configuration/channelb");
	ret[2] = m_container.at("channels/configuration/channelc");
	ret[3] = m_container.at("channels/configuration/channeld");
	for( std::vector< std::string >::iterator it = ret.begin(); it != ret.end(); ++it ){
		//FIXME: use regex
		if( it->size() < 10 || it->substr(0,7) != "Counter" || ( (*it)[8] != '1' && (*it)[8] != '2' ) || ( (*it)[9] != 'A' && (*it)[9] != 'B' ) ) throw std::logic_error( std::string("Bad value in channel configuration: ") + *it);
		*it = it->substr(8,2);
	}
	return ret;
}
	
}}
