/*
    $Id: params.h 180 2011-01-30 15:56:33Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _PARAMS_H
#define _PARAMS_H

#include <map>
#include <string>

#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "nocase_less.h"

namespace atmos {
namespace parameters {

using namespace boost::numeric::ublas;

class container {
public:
	typedef std::map< std::string, std::string, utils::algo::nocase_less< std::string > > storage_type;
	typedef storage_type::key_type   key_type;
	typedef storage_type::value_type value_type;
	typedef storage_type::const_iterator const_iterator;
private:
	storage_type m_storage;
public:
	container();
	~container();

	bool key_equal( const key_type& k1, const key_type& k2 );
	bool value_equal( const value_type& v1, const value_type& v2 );
	bool has(const storage_type::key_type& key) const;
	const storage_type::mapped_type& at(const storage_type::key_type& key) const;
	
	storage_type::mapped_type& operator[] (const storage_type::key_type& key) {
		return m_storage[key];
	}
	template<class Y> Y as(const storage_type::key_type& key) const {
		return boost::lexical_cast<Y>( at(key) );
	}
	template<class Y> Y as(const storage_type::key_type& key, const Y& def ) const {
		if( has( key ) )
			return boost::lexical_cast<Y>( at(key) );
		return def;
	}

	const_iterator begin() const { return m_storage.begin(); }
	const_iterator end() const { return m_storage.end(); }
};

typedef double               value_type;
typedef vector< value_type > vector_type;

class proxy {
private:
	const container& m_container;
public:
	proxy( const container& container );
	~proxy();

	std::string sitename() const { return m_container.at("general/site/sitename"); }
	value_type longitude() const;
	value_type latitude() const;
	yalpa::geographic< value_type > site() const;

	value_type accumtime()     const { return m_container.as< value_type >("operations/normal mode/accumtime"); }
	value_type basetime()      const { return m_container.as< value_type >("operations/normal mode/basetime"); }
	value_type exposition()    const { return m_container.as< value_type >("operations/normal mode/exposition"); }
	value_type magnification() const { return m_container.as< value_type >("optics/common/magnification"); }

	vector_type inner_diam()   const;
	vector_type outer_diam()   const;
	vector_type nonlinearity() const;
	vector_type nonpoisson() const;
	vector_type scatter() const;
	std::vector< std::string > counters() const;

	value_type dimm_aperture_base() const { return 0.01 * m_container.as< value_type >("general/dimm/aperturebase",20.0); }
	value_type dimm_aperture_size() const { return 0.01 * m_container.as< value_type >("general/dimm/aperturesize",10.0); }
	value_type dimm_scale() const { return 4.8481368e-06 * m_container.as< value_type >("camera/geometry/scale"); }
	value_type dimm_Ktransversal() const { return 0.358*(1.0*std::pow( dimm_aperture_size(), -1.0/3.0 ) - 0.811*std::pow( dimm_aperture_base(), -1.0/3.0 )); }
	value_type dimm_Klongitudinal() const { return 0.358*(1.0*std::pow( dimm_aperture_size(), -1.0/3.0 ) - 0.541*std::pow( dimm_aperture_base(), -1.0/3.0 )); }

};

}}

#endif // _PARAMS_H
