/*
    $Id: profile.cpp 327 2024-10-30 12:29:33Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <functional>

#include <boost/format.hpp>
#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/triangular.hpp>

#include <lsp/cholesky_decomposition.h>
#include <lsp/nnls.h>

#include "atmos.h"
#include "apply_to_all.h"
#include "covariance_decomposition.h"
#include "integrals.h"
#include "profile.h"
#include "utils.h"
#include "weif.h"

namespace atmos {
namespace mixture {
	
storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_mixture_column( *this, std::mem_fn(&detail::sample::mixture) ) {
}
storage::storage( const atmos::indices::storage& indices, const atmos::dimm::storage& dimm, double dimmscale ):
	table_storage< detail::sample >( indices.size() ),
	m_mixture_column( *this, std::mem_fn(&detail::sample::mixture) ) {
	using namespace boost::numeric::ublas;
	using namespace atmos::weight;
	
	if( dimm.size() > 1 ) {
		double dimmscale2 = dimmscale*dimmscale;
		dimm::storage::const_iterator dimm_current = dimm.begin();
		dimm::storage::const_iterator dimm_next = dimm_current + 1;
		indices::storage::const_iterator it = indices.begin();

		while( dimm_current != dimm.end() ) {
			while( it != indices.end() ) {
				if( dimm_next != dimm.end() && dimm_next->timestamp <= it->timestamp )
					break;
				
				push_back(repack_massdimm(as_vector(it->indices), dimmscale2*dimm_current->xvars(1), dimmscale2*dimm_current->yvars(1)));
				++it;
			}
			dimm_current = dimm_next;
			++dimm_next;
		}
	} else {
		for( indices::storage::const_iterator it = indices.begin(); it != indices.end(); ++it ) {
			push_back(as_vector(it->indices));
		}
	}
}

}
namespace profile {

functor::functor( const grid_type& grid, const mixture_covariance_type& cov_matrix, double secz, const weight_type& basewf ):
	m_grid( grid ),
	m_weight_matrix( utils::ublas::aetkin_weight_matrix(cov_matrix) ),
	m_invweight_matrix( solve( m_weight_matrix, identity_matrix< value_type >( m_weight_matrix.size1() ), lower_tag() ) ),
	m_secz( secz ) {
	using namespace utils::ublas;
	using namespace std;

	m_problem_matrix = prod( m_weight_matrix, subrange( vector_of_vector_to_matrix( project(grid*secz, basewf) ), 0, m_weight_matrix.size2(),0,grid.size()) );
}

functor::result_type functor::operator() ( const mixture_type& indices ) const {
	vector_type weighted_indices = prod( m_weight_matrix, indices );

	result_type ret;

	matrix_type cov = zero_matrix<double>( m_problem_matrix.size2(), m_problem_matrix.size2() );

	lsp::nnls< matrix_type, vector_type > nnls( m_problem_matrix, weighted_indices );
	nnls.solve( ret.cn2, cov );

	ret.residual = prod( m_problem_matrix, ret.cn2 ) - weighted_indices;

	ret.chi2 = inner_prod( ret.residual, ret.residual );

	ret.cn2 /= m_secz;
	return ret;
}
double functor::seeing( const cn2_type& cn2 ) const {
	return atmos::integrals::seeing( sum(cn2) );
}
vector_type functor::indices_residual( const residual_type& residual ) const {
	return prod( m_invweight_matrix, residual );
}

storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_cn2_column( *this, std::mem_fn(&detail::sample::cn2) ),
	m_residual_column( *this, std::mem_fn(&detail::sample::residual)  ),
	m_chi2_column( *this, std::mem_fn(&detail::sample::chi2) ) {
}

namespace filter {

chi2_gt_than::chi2_gt_than( const value_type& chi2 ): chi2_(chi2) {
}
bool chi2_gt_than::operator()( const sample_type& x ) const {
	return x.chi2 > chi2_;
}

}
}}
