/*
    $Id: profile.h 328 2024-10-30 12:37:15Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _PROFILE_H
#define _PROFILE_H

#include <functional>

#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

#include "indices.h"
#include "dimm.h"
#include "table_storage.h"

namespace atmos {
namespace mixture {
	using namespace boost::numeric::ublas;

	typedef double                 value_type;
	typedef vector<value_type>     vector_type;
	typedef vector_type::size_type size_type;
	typedef symmetric_matrix<value_type> symmetric_matrix_type;

	typedef vector_type           mixture_type;
	typedef symmetric_matrix_type mixture_covariance_type;

namespace detail {
	struct sample {
		typedef atmos::mixture::mixture_type mixture_type;

		mixture_type mixture;

		sample( const mixture_type& mixture ): mixture(mixture) {
		}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef storage self_type;
	typedef table_storage_column_getter< self_type, detail::sample::mixture_type > mixture_column_type;
private:
	mixture_column_type m_mixture_column;
public:
	storage( size_type reserve = 60 );
	storage( const atmos::indices::storage& indices, const atmos::dimm::storage& dimm, double dimmscale );

	const mixture_column_type& mixture() const { return m_mixture_column; }

	void push_back( const mixture_type& mixture ) {
		table_storage< detail::sample >::push_back( detail::sample( mixture ) );
	}
};
}
namespace profile {
	using namespace boost::numeric::ublas;

	typedef double                 value_type;
	typedef vector<value_type>     vector_type;
	typedef matrix<value_type>     matrix_type;
	typedef triangular_matrix<value_type> triangular_matrix_type;
	typedef vector_type::size_type size_type;

	typedef vector_type cn2_type;
	typedef vector_type residual_type;
	typedef value_type  chi2_type;

namespace detail {
	struct sample {
		typedef atmos::profile::cn2_type      cn2_type;
		typedef atmos::profile::residual_type residual_type;
		typedef atmos::profile::chi2_type     chi2_type;
	
		cn2_type       cn2;
		residual_type  residual;
		chi2_type      chi2;

		sample( const cn2_type& cn2, const residual_type& residual, const chi2_type& chi2 ):
			cn2(cn2), residual(residual), chi2(chi2) {
		}
		sample() {
		}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef storage self_type;
	typedef table_storage_column_getter< self_type, detail::sample::cn2_type >      cn2_column_type;
	typedef table_storage_column_getter< self_type, detail::sample::residual_type > residual_column_type;
	typedef table_storage_column_getter< self_type, detail::sample::chi2_type >     chi2_column_type;
private:
	cn2_column_type      m_cn2_column;
	residual_column_type m_residual_column;
	chi2_column_type     m_chi2_column;
public:
	storage( size_type reserve = 60 );

	const cn2_column_type&      cn2()      const { return m_cn2_column; }
	const residual_column_type& residual() const { return m_residual_column; }
	const chi2_column_type&     chi2()     const { return m_chi2_column; }

	void push_back( const detail::sample& profile ) {
		table_storage< detail::sample >::push_back( profile );
	}
};

class functor {
	public:
		typedef detail::sample result_type;
		typedef atmos::mixture::mixture_type mixture_type;
		typedef atmos::mixture::mixture_covariance_type mixture_covariance_type;

		typedef vector_type grid_type;
		typedef std::function< vector<double> (double) > weight_type;
	private:
		grid_type     m_grid;
		triangular_matrix_type m_weight_matrix;
		triangular_matrix_type m_invweight_matrix;
		matrix_type   m_problem_matrix;
		double        m_secz;

	public:
		functor( const grid_type& grid, const mixture_covariance_type& cov_matrix, double secz, const weight_type& basewf );

		result_type operator() ( const mixture_type& indices ) const;

		double seeing( const cn2_type& cn2 ) const;
		vector_type indices_residual( const residual_type& residual ) const;
};

namespace filter {
	class chi2_gt_than {
	public:
		typedef atmos::profile::detail::sample sample_type;
		typedef atmos::profile::value_type value_type;
	private:
		value_type chi2_;
	public:
		chi2_gt_than( const value_type& chi2 );
		bool operator()( const sample_type& x ) const;
	};
}
}}

#endif // _PROFILE_H
