/*
    $Id: spline.h 287 2012-10-08 19:06:37Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _SPLINE_H
#define _SPLINE_H

#include <boost/numeric/ublas/vector.hpp>

namespace utils {
namespace ublas {

using namespace boost::numeric::ublas;

template<class XT, class YT, class CXT, class CYT> class cubic_spline;

namespace detail {
	template< class XT, class YT, class CXT, class CYT, class E > class spline_vector_result:
		public vector_expression< spline_vector_result<XT,YT,CXT,CYT,E> > {
	public:
		typedef cubic_spline<XT,YT,CXT,CYT> cubic_spline_type;
		typedef const cubic_spline_type& spline_const_reference;
		typedef typename vector_expression< E >::expression_type::size_type size_type;
		typedef typename cubic_spline_type::second_value_type value_type;
		typedef void const_iterator;

	public:
		spline_vector_result( const cubic_spline_type& spline, const vector_expression<E>& e ): spline_(spline), e_(e) {
		}
		size_type size() const { return e_().size(); }

		template<class V> void apply( V& v ) const {
			typedef typename V::size_type size_type;
			typedef typename cubic_spline_type::first_size_type first_size_type;

			size_type size (BOOST_UBLAS_SAME(v.size (), e_().size ()));

			for( size_type i = 0; i < size; ++i ){
				v(i) = spline_( e_()(i) );
			}
		}
	private:
		spline_const_reference spline_;
		const vector_expression<E>& e_;
	};

	template<template <class T1, class T2> class F, class V, class XT, class YT, class CXT, class CYT, class E>
	void vector_assign (V &v, const vector_expression< spline_vector_result<XT, YT, CXT, CYT, E> > &e) {
		e().apply( v );
	}
}

template<class XT, class YT, class CXT = typename XT::const_closure_type, class CYT = typename YT::const_closure_type > class cubic_spline {
public:
	typedef XT first_type;
	typedef YT second_type;

	typedef cubic_spline<XT,YT> self_type;

	typedef CXT first_closure_type;
	typedef CYT second_closure_type;

	typedef typename first_type::value_type  first_value_type;
	typedef typename first_type::size_type   first_size_type;
	typedef typename second_type::value_type second_value_type;
	typedef typename second_type::size_type  second_size_type;

	typedef vector< second_value_type > storage_vector_type;
public:
	template<class T> class boundary {
	public:
		typedef T value_type;
	public:
		boundary(): natural_(true) {
		}
		boundary( const value_type& value ): value_(value), natural_(false) {
		}

		const bool natural() const { return natural_; }
		operator value_type() const { return value_; }
	private:
		value_type value_;
		bool natural_;
	};

	typedef boundary< second_value_type > boundary_type;

public:
	cubic_spline(
		const XT& first,
		const YT& second,
		const boundary_type& lb = boundary_type(),
		const boundary_type& hb = boundary_type() ):
		first_( first ), second_( second ), d2_( first.size() ), m_ihint(0) {

		assert( first.size() == second.size() );
		update( lb, hb );
	}
	cubic_spline(
		const std::pair< XT, YT > pair,
		const boundary_type& lb = boundary_type(),
		const boundary_type& hb = boundary_type() ):
		first_( pair.first ), second_( pair.second ), d2_( first_.size() ), m_ihint(0) {

		assert( first_.size() == second_.size() );
		update( lb, hb );
	}

	const first_closure_type&  first_data() const  { return first_; }
	const second_closure_type& second_data() const { return second_; }

	first_size_type size() const { return first_.size(); }
	bool inrange( const second_value_type& x ) const { return x >= first_(0) && x < first_( first_.size() - 1 ); }

	void update( const boundary_type& lb = boundary_type(), const boundary_type& hb = boundary_type() ) {
		second_size_type n = second_.size();
		vector< first_value_type > u( n );
		second_size_type i;

		/* boundary */
		if( lb.natural() ) {
			d2_(0) = second_(0) * 0.0;
			u(0) = first_value_type( 0 );
		} else {
			d2_(0) = (first_value_type(3)/(first_(1)-first_(0))) * ((second_(1)-second_(0)) / (first_(1)-first_(0))-second_value_type(lb));
			u(0) = first_value_type(-0.5);
		}

		/* forward sweep */
		for( i = 1; i < n-1; ++i ) {
			first_value_type sig = (first_(i)-first_(i-1)) / (first_(i+1)-first_(i-1));
			first_value_type p  = sig * u( i-1 ) + first_value_type(2);
			u(i) = ( sig - first_value_type(1) ) / p;
			d2_(i) = (second_(i+1) - second_(i)) / (first_(i+1)-first_(i)) - (second_(i) - second_(i-1)) / (first_(i)-first_(i-1));
			d2_(i) = (first_value_type(6)*d2_(i)/(first_(i+1)-first_(i-1))-sig*d2_(i-1)) / p;
		}

		/* boundary */
		second_value_type qn;
		if( hb.natural() ) {
			u(i-1) = first_value_type(0);
			qn = second_(0) * 0.0;
		} else {
			u(i-1) = first_value_type(0.5);
			qn = (first_value_type(3)/(first_(n-1)-first_(n-2)))*(second_value_type(hb)-(second_(n-1)-second_(n-2)) / (first_(n-1)-first_(n-2)));
		}

		/* back substitution */
		d2_(n-1) = ( qn - u(i-1) * d2_(n-2) ) / ( u(i-1) * u(n-2) + first_value_type(1) );
		for( i = n-1; i > 0; --i ) {
			d2_(i-1) = d2_(i-1) + d2_(i) * u(i-1);
		}
	}

	second_value_type operator() ( const first_value_type& x ) const {
		m_ihint = hunt( x, m_ihint );
		return eval( x, m_ihint );
	}
	template<class E> detail::spline_vector_result< XT, YT, CXT, CYT, E > operator() (const vector_expression<E> &e) const {
		return detail::spline_vector_result< XT, YT, CXT, CYT, E >( *this, e );
	}
private:
	second_value_type eval( const first_value_type& x, const first_size_type& node ) const {
		const first_value_type h = first_( node + 1 ) - first_( node );
		const first_value_type a = ( first_( node + 1 ) - x ) / h;
		const first_value_type b = ( x - first_( node ) ) / h;
		return a*second_( node ) +
			b*second_( node + 1 ) +
			((a*a*a-a)*d2_( node ) + (b*b*b-b)*d2_( node + 1 ))*(h*h)/6.0;
	}
	first_size_type hunt( const first_value_type& x, const first_size_type& ii ) const {
		first_value_type ix = first_(ii);

		first_size_type li = 0;
		first_size_type hi = first_.size();
		first_size_type mi = ii;
		first_size_type step = 1;

		if( x > ix ) {
			do {
				if( hi - mi < step ) {
					mi = hi;
					break;
				}
				li = mi;
				mi += step;
				step += step;
			} while ( x > first_(mi) );
			hi = mi;
		} else {
			do {
				if( mi - li < step ) {
					mi = li;
					break;
				}
				hi = mi;
				mi -= step;
				step += step;
			} while ( x < first_(mi) );
			li = mi;
		}

		while( hi > li + 1 ) {
			mi = (hi + li) / 2;
			if( x > first_(mi) ) {
				li = mi;
			} else {
				hi = mi;
			}
		}

		if( li >= first_.size() - 1 )
			return first_.size() - 2;
		return li;
	}

private:
	first_closure_type  first_;
	second_closure_type second_;
	storage_vector_type d2_;

	mutable first_size_type m_ihint;
};

template<class XT, class YT> cubic_spline<XT,YT> make_cubic_spline( const XT& x, const YT& y ){
	return cubic_spline<XT,YT>(x,y);
}
template<class XT, class YT> cubic_spline<XT,YT> make_cubic_spline( const XT& x, const YT& y, const typename cubic_spline<XT,YT>::boundary_type& yp1 ){
	return cubic_spline<XT,YT>(x,y,yp1);
}
template<class XT, class YT> cubic_spline<XT,YT> make_cubic_spline( const XT& x, const YT& y, const typename cubic_spline<XT,YT>::boundary_type& yp1, const typename cubic_spline<XT,YT>::boundary_type& yp2 ){
	return cubic_spline<XT,YT>(x,y,yp1,yp2);
}

}
}

#endif // _SPLINE_H
