/*
    $Id: ublas_math.h 281 2012-02-23 16:14:32Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _UBLAS_MATH_H
#define _UBLAS_MATH_H

#include <cmath>

#include "apply_to_all.h"
#include "lambert.h"

namespace utils {
namespace ublas {

namespace detail {
	template <class T> class inv {
	public:
		typedef T value_type;
		typedef T result_type;
		static result_type apply( const value_type& x) {
			return value_type(1) / x;
		}
	};
	template <class T> class exp {
	public:
		typedef T value_type;
		typedef T result_type;
		static result_type apply( const value_type& x) {
			return std::exp( x );
		}
	};
	template <class T> class log {
	public:
		typedef T value_type;
		typedef T result_type;
		static result_type apply( const value_type& x) {
			return std::log( x );
		}
	};
	template <class T> class lambert_W0 {
	public:
		typedef T value_type;
		typedef T result_type;
		static result_type apply( const value_type& x) {
			return utils::math::lambert_W0(x);
		}
	};
	template <class T> class sqrt {
	public:
		typedef T value_type;
		typedef T result_type;
		static result_type apply( const value_type& x) {
			return std::sqrt(x);
		}
	};
	template <class T> class xp1 {
	public:
		typedef T value_type;
		typedef T result_type;
		static result_type apply( const value_type& x) {
			return x + value_type(1);
		}
	};
	template <class T> class abs {
	public:
		typedef T value_type;
		typedef T result_type;
		static result_type apply( const value_type& x) {
			return std::abs(x);
		}
	};
	
	template<class T, class U> struct scalar_pow: public scalar_binary_functor<T, U> {
		typedef typename scalar_binary_functor<T, U>::argument1_type argument1_type;
		typedef typename scalar_binary_functor<T, U>::argument2_type argument2_type;
		typedef typename boost::remove_reference< argument1_type >::type result_type;

		static BOOST_UBLAS_INLINE
		result_type apply (argument1_type t, argument2_type p) {
			using namespace std;
			// we'll find either std::pow or else another version via ADL:
			return pow (t,p);
		}
	};
	
}
	
using namespace boost::numeric::ublas;

template<class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, detail::inv< typename E::value_type > >::result_type inv(const vector_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::inv< value_type >() );
}
template<class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, detail::inv< typename E::value_type > >::result_type inv(const matrix_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::inv< value_type >() );
}

template<class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, detail::exp< typename E::value_type > >::result_type exp(const vector_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::exp< value_type >() );
}
template<class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, detail::exp< typename E::value_type > >::result_type exp(const matrix_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::exp< value_type >() );
}

template<class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, detail::log< typename E::value_type > >::result_type log(const vector_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::log< value_type >() );
}
template<class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, detail::log< typename E::value_type > >::result_type log(const matrix_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::log< value_type >() );
}

template<class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, detail::sqrt< typename E::value_type > >::result_type sqrt(const vector_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::sqrt< value_type >() );
}
template<class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, detail::sqrt< typename E::value_type > >::result_type sqrt(const matrix_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::sqrt< value_type >() );
}

template<class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, detail::xp1< typename E::value_type > >::result_type xp1(const vector_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::xp1< value_type >() );
}
template<class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, detail::xp1< typename E::value_type > >::result_type xp1(const matrix_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::xp1< value_type >() );
}

template<class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, detail::lambert_W0< typename E::value_type > >::result_type lambert_W0(const vector_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::lambert_W0< value_type >() );
}
template<class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, detail::lambert_W0< typename E::value_type > >::result_type lambert_W0(const matrix_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::lambert_W0< value_type >() );
}

template<class E> BOOST_UBLAS_INLINE typename vector_unary_traits<E, detail::abs< typename E::value_type > >::result_type abs(const vector_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::abs< value_type >() );
}
template<class E> BOOST_UBLAS_INLINE typename matrix_unary1_traits<E, detail::abs< typename E::value_type > >::result_type abs(const matrix_expression<E> &e ) {
	typedef typename E::value_type value_type;
	return apply_to_all( e, detail::abs< value_type >() );
}

// pow (v1, v2) [i] = pow( v1 [i], v2 [i] )
template<class E1, class E2> BOOST_UBLAS_INLINE
typename vector_binary_traits<E1, E2, detail::scalar_pow<typename E1::value_type, typename E2::value_type> >::result_type
element_pow (const vector_expression<E1> &e1, const vector_expression<E2> &e2) {
	typedef typename vector_binary_traits<E1, E2, detail::scalar_pow<typename E1::value_type, typename E2::value_type> >::expression_type expression_type;
	return expression_type (e1 (), e2 ());
}
// pow (m1, m2) [i] [j] = pow (m1 [i] [j], m2 [i] [j])
template<class E1, class E2> BOOST_UBLAS_INLINE
typename matrix_binary_traits<E1, E2, detail::scalar_pow<typename E1::value_type, typename E2::value_type> >::result_type
element_pow (const matrix_expression<E1> &e1, const matrix_expression<E2> &e2) {
	typedef typename matrix_binary_traits<E1, E2, detail::scalar_pow<typename E1::value_type, typename E2::value_type> >::expression_type expression_type;
	return expression_type (e1 (), e2 ());
}

// pow (v, t) [i] = pow( v [i], t )
template<class E1, class T2> BOOST_UBLAS_INLINE
typename vector_binary_scalar2_traits<E1, const T2, detail::scalar_pow<typename E1::value_type, T2> >::result_type
pow (const vector_expression<E1> &e1, const T2 &e2) {
	typedef typename vector_binary_scalar2_traits<E1, const T2, detail::scalar_pow<typename E1::value_type, T2> >::expression_type expression_type;
	return expression_type (e1 (), e2);
}
// pow (m, p) [i] [j] = pow (m [i] [j], p)
template<class E1, class T2> BOOST_UBLAS_INLINE
typename matrix_binary_scalar2_traits<E1, const T2, detail::scalar_pow<typename E1::value_type, T2> >::result_type
pow (const matrix_expression<E1> &e1, const T2 &e2) {
	typedef typename matrix_binary_scalar2_traits<E1, const T2, detail::scalar_pow<typename E1::value_type, T2> >::expression_type expression_type;
	return expression_type (e1 (), e2);
}

}}

#endif // _UBLAS_MATH_H
