/*
    $Id: weif.cpp 328 2024-10-30 12:37:15Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <functional>

#include <boost/numeric/ublas/vector_proxy.hpp>

#include "atmos.h"
#include "fft.h"
#include "gsl_wrap.h"
#include "utils.h"
#include "weif.h"

namespace atmos {
namespace weight {

const double POWIND = -2.66666666667;
const double MKM2CM = 1E-4;
const double EPS = 1.0e-4 ;
const double CALIBR  = 44.62621363;
const double DIMMCONST = 2.403*4.64159;
const double KM2CM  = 1E5 ;
const int  WORKSPACE = 10000;

static gsl::set_error_handler_off set_error_handler_off_;

struct mass_params {
	typedef std::function< std::complex<double>(double) > sfun_type;

	double diameter1;
	double diameter2;
	double diameter3;
	double diameter4;
	double PIx2lambda0;
	sfun_type sfun;
	double altitude;
	double powind;

	mass_params( const sfun_type& sfun ): sfun(sfun) {}
};

struct dimm_params {
	typedef std::function< std::complex<double>(double) > sfun_type;

	double aperture;
	double dimmbase;
	double PIx2lambda0;
	double cos2phi;
	sfun_type sfun;
	double altitude;

	dimm_params( const sfun_type& sfun ): sfun(sfun) {}
};

double mass_integrand(double x, void * params) {
	using namespace std;

	const mass_params* p = (mass_params*)params;

	if( p->diameter1 == p->diameter2 || p->diameter3 == p->diameter4 ) { return 0.0; }

	const double f = p->altitude*x*x / 2.0;
	const std::complex<double>& sfun_f = p->sfun(f);
	double sf = imag( sfun_f ) * cos( p->PIx2lambda0*f ) + real( sfun_f ) * sin( p->PIx2lambda0*f );
	sf = sf*sf*pow( x, p->powind );
	double p1 = ( p->diameter1*j1( p->diameter1*x ) - p->diameter2*j1( p->diameter2*x ) ) / ( p->diameter1*p->diameter1 - p->diameter2*p->diameter2 );
	double p3 = ( p->diameter3*j1( p->diameter3*x ) - p->diameter4*j1( p->diameter4*x ) ) / ( p->diameter3*p->diameter3 - p->diameter4*p->diameter4 );

	return sf * 4.0 * p1 * p3 / (x*x);
}

double dimm_integrand(double x, void * params) {
	const double powdiff = -0.666666666667;

	using namespace std;

	const dimm_params* p = (dimm_params*)params;

	const double f = p->altitude*x*x / 2.0;
	const double a = p->aperture*x;
	const double b = p->dimmbase*x;
	const std::complex<double>& sfun_f = p->sfun(f);
	double cf = real( sfun_f ) * cos( p->PIx2lambda0*f ) - imag( sfun_f ) * sin( p->PIx2lambda0*f );
	cf = cf*( a != 0.0 ? 8.0*jn(2,a)/(a*a) : 1.0 );
	cf = cf*cf*pow( x, powdiff );

	return cf*( 1.0 - j0(b) + p->cos2phi*jn(2,b) );
}

size_t next_pow2( size_t x ) {
	if( !(x&(x-1)) ) return x;
	for( size_t i = 1; i < std::numeric_limits<size_t>::digits; i<<=1 ) {
		x |= (x >> i);
	}
	return (x & ~(x >> 1)) << 1;
}

template<class V,class V2> typename V::const_iterator lambda_effective( const V& lambda, const V2& dist ) {
	typedef typename V::value_type value_type;
	typedef typename V::const_iterator const_iterator;

	value_type leff = inner_prod( lambda, dist ) / sum( dist );
	const_iterator it = std::lower_bound( lambda.begin(), lambda.end(), leff );

	return it;
}

std::pair< vector_type, vector< complex_type > > spectral_band_fourier( const vector<double>& lambda, const vector<double>& dist, unsigned int center ) {
	typedef vector<double>::size_type size_type;

	const size_type grid_factor = 4;
	const double dlambda = lambda(1) - lambda(0);
	const size_type distft_size = 2 * grid_factor * next_pow2( dist.size() );

	vector< std::complex<double> > dist_ft = zero_vector< std::complex<double> >( distft_size );
	for( vector<double>::size_type l = 0; l < dist.size(); l++ ) {
		vector<double>::size_type j = ( l < center ) ? distft_size + l - center : l - center;
		dist_ft(j) = dist(l);
	}
	noalias( dist_ft ) = utils::ublas::fft_radix2( dist_ft );

	vector< double > freq( distft_size/2 );
	for( vector<double>::size_type i = 0; i < freq.size(); ++i ) {
		freq(i) = i/(dlambda*MKM2CM)/(distft_size);
	}

	return std::make_pair( freq, subrange( dist_ft, 0, distft_size/2 ) );
}

namespace detail {

basic_weight::value_type basic_weight::do_calculate( value_type height, value_type eps1, value_type eps2, value_type eps3, value_type eps4 ) const {
	gsl::integration_workspace w(WORKSPACE);
	double error;
	mass_params params( mass_sfourier_ );
	gsl_function F;
	F.function = &mass_integrand;
	F.params = &params;
	params.diameter1 = M_PI*eps1;
	params.diameter2 = M_PI*eps2;
	params.diameter3 = M_PI*eps3;
	params.diameter4 = M_PI*eps4;
	params.PIx2lambda0 = 2*M_PI * mass_leff_ * MKM2CM;
	params.powind = POWIND + freq_power_;
	double result;
	params.altitude = height*KM2CM;
	double hi_limit = std::sqrt( 2 * mass_freq_max_ / params.altitude );
	if( std::isfinite( hi_limit ) )
		gsl::integration_qags( &F, 0, hi_limit, 0, EPS, WORKSPACE, w, &result, &error);
	else
		result = 0.0;
	return result*CALIBR*std::pow(M_PI*100,freq_power_);
}

basic_weight::basic_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& freq_power ):
	ap_inn_( ap_inn ),
	ap_out_( ap_out ),
	mass_leff_index_( lambda_effective( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ) ).index() ),
	mass_leff_( mass_lambda( mass_leff_index_ ) ),
	mass_sfourier_( spectral_band_fourier( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ), mass_leff_index_ ) ),
	freq_power_(freq_power) {

	mass_freq_max_ = mass_sfourier_.first_data()( mass_sfourier_.size() - 1 );
}

basic_weight::result_type basic_weight::operator() ( const arg_type& height ) const {
	indices_type ret( atmos::channels, atmos::channels );

	for( indices_type::size_type i = 0; i < BOOST_UBLAS_SAME(ret.size1(),ret.size2()); ++i ) {
		for( indices_type::size_type j = i; j < BOOST_UBLAS_SAME(ret.size1(),ret.size2()); ++j ) {
			ret(i,j) = do_calculate( height, ap_out_(i), ap_inn_(i), ap_out_(j), ap_inn_(j) );
		}
	}

	return as_vector(ret);
}

}

massdimm_weight::value_type massdimm_weight::do_dimm_calculate( value_type height, value_type diam, value_type dbase, value_type cos2phi ) const {
	gsl::integration_workspace w(WORKSPACE);
	double error;
	dimm_params params(dimm_sfourier_);
	gsl_function F;
	F.function = &dimm_integrand;
	F.params = &params;
	params.aperture = M_PI*diam;
	params.dimmbase = 2*M_PI*dbase;
	params.PIx2lambda0 = 2*M_PI*dimm_leff_*MKM2CM;
	params.cos2phi = cos2phi;
	double result;
	params.altitude = height*KM2CM;
	double hi_limit = std::sqrt( 2 * dimm_freq_max_ / params.altitude );
	if( std::isfinite( hi_limit ) )
		gsl::integration_qags( &F, 0, hi_limit, 0, EPS, WORKSPACE, w, &result, &error);
	else
		gsl::integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result, &error);
	return result*DIMMCONST;
}

massdimm_weight::massdimm_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& dimm_lambda, const vector_type& dimm_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& dimmsize, const value_type& dimmbase ):
	basic_weight( mass_lambda, mass_response, ap_inn, ap_out ),
	dimmsize_( dimmsize * 100 ), dimmbase_( dimmbase * 100 ),
	dimm_leff_index_( lambda_effective( dimm_lambda, dimm_response ).index() ),
	dimm_leff_( dimm_lambda( dimm_leff_index_ ) ),
	dimm_sfourier_( spectral_band_fourier( dimm_lambda, dimm_response, dimm_leff_index_ ) ) {

	dimm_freq_max_ = dimm_sfourier_.first_data()( dimm_sfourier_.size() - 1 );
}

massdimm_weight::result_type massdimm_weight::operator() ( const arg_type& height ) const {
	return repack_massdimm( basic_weight::operator()(height), 
		do_dimm_calculate( height, dimmsize_, dimmbase_, 1.0 ),   // longitudinal, phi = 0.0
		do_dimm_calculate( height, dimmsize_, dimmbase_,-1.0 ) ); // transversal,  phi = Pi/2
}

mass_shortexposure_weight::mass_shortexposure_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out ):
	basic_weight( mass_lambda, mass_response, ap_inn, ap_out, 2.0 ) {
}

}}
