/*
    $Id: weif.h 279 2012-02-23 14:27:56Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _WEIF_H
#define _WEIF_H

#include <complex>

#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "atmos.h"
#include "spline.h"

namespace atmos {
namespace weight {
	using namespace boost::numeric::ublas;

	typedef double value_type;
	typedef std::complex< value_type > complex_type;
	typedef symmetric_matrix< value_type > symmetric_matrix_type;
	typedef vector< value_type >   vector_type;
	typedef vector_type::size_type size_type;

	typedef symmetric_matrix_type  indices_type;
	typedef vector_type            weight_type;

#ifdef BOOST_UBLAS_SHALLOW_ARRAY_ADAPTOR
#  define AS_VECTOR__ARRAY_TYPE shallow_array_adaptor<const T>
#  define AS_VECTOR__ARRAY_CONSTRUCT(size,mtx) size, mtx.data().begin()
#else
#  warning "Please, consider using -DBOOST_UBLAS_SHALLOW_ARRAY_ADAPTOR option."
#  define AS_VECTOR__ARRAY_TYPE typename vector<T>::array_type
#  define AS_VECTOR__ARRAY_CONSTRUCT(size,mtx) mtx.data()
#endif // BOOST_UBLAS_SHALLOW_ARRAY_ADAPTOR
template<class T, class TRI, class L, class A> vector<T, AS_VECTOR__ARRAY_TYPE > as_vector( const symmetric_matrix<T,TRI,L,A>& mtx ) {
	typedef typename symmetric_matrix<T,TRI,L,A>::size_type size_type;
	typedef AS_VECTOR__ARRAY_TYPE array_type;

	const size_type size = L::triangular_size(mtx.size1(), mtx.size2());

	return vector<T, AS_VECTOR__ARRAY_TYPE >(size, array_type(AS_VECTOR__ARRAY_CONSTRUCT(size,mtx)));
}
template<class T, class A> symmetric_matrix< T, lower, row_major, AS_VECTOR__ARRAY_TYPE > as_symmetric( const vector<T,A>& vec ) {
	typedef typename vector<T,A>::size_type size_type;
	typedef AS_VECTOR__ARRAY_TYPE array_type;
	typedef row_major layout_type;

	const size_type n = 8*vec.size()+1;
	size_type size = 4*vec.size()+1, size0;
	do { size0 = size; size = (size+n/size)/2; } while( size != size0 );
	size = (size-1)/2;

	return symmetric_matrix< T, lower, row_major, AS_VECTOR__ARRAY_TYPE >(size, array_type(AS_VECTOR__ARRAY_CONSTRUCT(layout_type::triangular_size(size,size),vec)));
}
template<class T, class A> vector<T> repack_massdimm( const vector<T, A>& mass_indices, const T& dimm_long, const T& dimm_trans ) {
	typedef T value_type;
	typedef typename vector<T>::size_type size_type;

	const value_type scale_factor = 1e+11;
	const size_type  size = mass_indices.size() + 2;

	vector<T> ret(size);
	for( size_type i = 0; i < mass_indices.size(); ++i ) { ret(i) = mass_indices(i); }
	ret(size-2) = dimm_long  * scale_factor;
	ret(size-1) = dimm_trans * scale_factor;
	return ret;
}
template<class T, class A> vector<T, AS_VECTOR__ARRAY_TYPE > unpack_mass( const vector<T, A>& indices ) {
	typedef typename vector<T, A>::size_type size_type;
	typedef AS_VECTOR__ARRAY_TYPE array_type;
	typedef T value_type;

	const size_type size = atmos::indices_size;

	assert( indices.size() >= size );

	return vector<T, AS_VECTOR__ARRAY_TYPE >(size, array_type(AS_VECTOR__ARRAY_CONSTRUCT(size,indices)));
}
template<class T, class A> std::pair<T,T> unpack_dimm( const vector<T, A>& indices ) {
	typedef typename vector<T, A>::size_type size_type;

	const size_type size = indices.size();

	assert( size >= atmos::indices_size + 2 );

	return std::make_pair(indices(size-2),indices(size-1));
}
#undef AS_VECTOR__ARRAY_TYPE
#undef AS_VECTOR__ARRAY_CONSTRUCT

namespace detail {
	class basic_weight {
	public:
		typedef atmos::weight::value_type value_type;
		typedef value_type  arg_type;
		typedef weight_type result_type;

	protected:
		typedef utils::ublas::cubic_spline< vector_type, vector< complex_type >, vector_type, vector< complex_type > > sfourier_spline_type;
		typedef sfourier_spline_type sfourier_type;
	private:
		value_type do_calculate( value_type height, value_type eps1, value_type eps2, value_type eps3, value_type eps4 ) const;
	public:
		basic_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& freq_power = value_type(0) );

		virtual result_type operator() ( const arg_type& height ) const;
	private:
		vector_type ap_inn_;
		vector_type ap_out_;
		size_type mass_leff_index_;
		value_type mass_leff_;
		sfourier_type mass_sfourier_;
		value_type mass_freq_max_;
		value_type freq_power_;
	};
}

class massdimm_weight:
	public detail::basic_weight {
private:
	value_type do_dimm_calculate( value_type height, value_type diam, value_type dbase, value_type cos2phi ) const;
public:
	massdimm_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& dimm_lambda, const vector_type& dimm_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& dimmsize, const value_type& dimmbase );
	
	result_type operator() ( const arg_type& height ) const;
private:
	value_type dimmsize_;
	value_type dimmbase_;
	size_type dimm_leff_index_;
	value_type dimm_leff_;
	sfourier_type dimm_sfourier_;
	value_type dimm_freq_max_;
};

class mass_shortexposure_weight:
	public detail::basic_weight {
public:
	mass_shortexposure_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out );
};
}}

#endif // _WEIF_H
