/*
    $Id: worksheet.cpp 329 2024-10-30 12:52:28Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath> // for std::cos
#include <iostream>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include <yalpa/sun.h>
#include <yalpa/sidtime.h>
#include <yalpa/spheric.h>

#include "atmos.h"
#include "average.h"
#include "dimm.h"
#include "weif.h" // for wf::*_generator
#include "worksheet.h"

namespace atmos {

struct worksheet::profile_grid {
	boost::numeric::ublas::vector<double> vec;

	profile_grid( const boost::numeric::ublas::vector<double>& vec ): vec(vec) {}
};

class worksheet::object:
	public local_context< worksheet > {
public:
	typedef context_type::input::time_type time_type;
	typedef context_type::weight_factory_type::nullexposure_weight_type nullexposure_weight_type;
	typedef context_type::weight_factory_type::shortexposure_weight_type shortexposure_weight_type;
private:
	std::string m_spectral;
	yalpa::equatorial<double> m_coords;
public:
	object( const yalpa::equatorial<double> ecoords, const std::string& spectral, context_reference_type ctx ): local_context_type( ctx ), m_spectral(spectral), m_coords(ecoords) {}

	double airmass( const time_type& time ) const;
	const nullexposure_weight_type& weight();
	const shortexposure_weight_type& short_weight();
	const std::string& sclass() const { return m_spectral; }
};

worksheet::input::measurement_accumulator::~measurement_accumulator() {
}

class worksheet::input::background_accumulator:
	public measurement_accumulator {
private:
	std::unique_ptr< atmos::background::model::abstract > m_model;
	const atmos::parameters::vector_type m_deadtime;
	const double m_exposition;	
	atmos::background::filter::nonpoisson m_nonpoisson_filter;
public:
	background_accumulator( const time_type& time, context_reference_type ctx ):
		measurement_accumulator( time, ctx ),
		m_model( ctx.m_background_factory.create_new_model(time) ),
		m_deadtime( ctx.parameters().nonlinearity() ),
		m_exposition( ctx.parameters().exposition() ),
		m_nonpoisson_filter(
			ctx.m_va["background-threshold"].as<double>(),
			ctx.parameters().nonpoisson()(3),
			1000 * ctx.parameters().basetime() / ctx.parameters().exposition() ) {

		if( !m_model.get() )
			throw std::logic_error("There is not any background model");
	};
	virtual ~background_accumulator() {};
	virtual void accumulate(  const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
		using atmos::indices::flux_nonlinearity_correction;
		
		if( ! m_nonpoisson_filter( time, mean, vars ) ) {
			if( m_model.get() ) {
				m_model->accumulate( time, flux_nonlinearity_correction( mean / m_exposition, m_deadtime ) );
			}
		}
	}
	virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {
	}
	virtual void flush() {
		context().m_file_output.background( starttime() );

		if( m_model.get() && m_model->size() > 0 ) {
			context().m_background_factory.handle_new_model( m_model.release() );
			context().m_file_output.fluxes( starttime(), context().m_background_factory.background( starttime() ) );
		}
	}
};
class worksheet::input::normal_accumulator:
	public measurement_accumulator {
public:
	typedef context_type::input::size_type size_type;
private:
	time_type m_accumulation_begin;
	time_type m_accumulation_end;
	double m_mean_mz;
	size_type m_num;

	atmos::dimm::storage    m_dimm;
	atmos::indices::storage m_indices;
	
	//FIXME: use params proxy lazy initialization
	const atmos::parameters::vector_type m_deadtime;
	const atmos::parameters::vector_type m_scatter;
	const atmos::parameters::vector_type m_npoisson;
	const double m_exposition;	

	atmos::indices::filter::flux_threshold m_flux_filter;
public:
	normal_accumulator( const time_type& time, context_reference_type ctx ):
		measurement_accumulator( time, ctx ),
		m_mean_mz(0.0),
		m_num(0),
		m_deadtime( ctx.parameters().nonlinearity() ),
		m_scatter( ctx.parameters().scatter() ),
		m_npoisson( ctx.parameters().nonpoisson() ),
		m_exposition( ctx.parameters().exposition() ),
		m_flux_filter( ctx.m_va["flux-threshold"].as<double>() ) {

	}
	virtual ~normal_accumulator() {};
	virtual void accumulate( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
		/* mean is in photons
		   flux is in photons/msec */
		using namespace atmos::indices;
		
		if( m_accumulation_begin == boost::posix_time::not_a_date_time )
			m_accumulation_begin = time;
		m_accumulation_end = time;

		m_num++;
		m_mean_mz = m_mean_mz * ( double(m_num-1) / double(m_num) ) + context().curobject().airmass( time ) * ( 1.0/m_num );

		flux_type flux = mean / m_exposition;
		flux_type flux_corrected = flux_nonlinearity_correction( flux, m_deadtime );

		const std::tuple< vars_type, rho1_type, rho2_type >& vars_corrected = vars_nonlinearity_correction(
			flux_corrected,
			element_prod( vars, outer_prod( flux, flux ) ),
			element_prod( rho1, outer_prod( flux, flux ) ),
			element_prod( rho2, element_prod( flux, flux ) ), m_deadtime, m_npoisson, m_exposition );

		flux_corrected = flux_background_correction( flux_corrected, m_scatter, context().background_factory().background( time ) );

		if( ! m_flux_filter( flux_corrected ) ) {
			const vars_type& vars_nullexp = vars_exposure_correction( flux_corrected, std::get<0>(vars_corrected), std::get<1>(vars_corrected) );
			
			const indesis_type& indesis         = calculate_indesis( flux_corrected, std::get<0>(vars_corrected), std::get<1>(vars_corrected), std::get<2>(vars_corrected) );
			const indices_type& indecis_nullexp = calculate_indices( flux_corrected, vars_nullexp );
			const indices_type& indecis         = calculate_indices( flux_corrected, std::get<0>(vars_corrected) );

			m_indices.push_back( time, flux_corrected, indecis_nullexp, indesis, indecis - calculate_indices(flux_corrected, std::get<1>(vars_corrected)) );
		}
	}
	virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {
		using atmos::dimm::correction;
		
		if( m_accumulation_begin == boost::posix_time::not_a_date_time )
			m_accumulation_begin = time;
		m_accumulation_end = time;

		m_dimm.push_back( time, framenum, left, right, correction( xvars ), correction( yvars ) );
	}
	virtual void flush() {
		using namespace utils::algo;
		using namespace atmos;
		using namespace boost::numeric::ublas;

		const boost::posix_time::ptime& meantime = m_accumulation_begin + (m_accumulation_end-m_accumulation_begin)/2;

		typedef context_type::integrals_factory_type integrals_factory_type;
		typedef integrals_factory_type::value_type   integrals_functor_type;
		typedef integrals_functor_type::result_type  integrals_result_type;
		typedef profile::functor                    restoration_functor_type;
		typedef restoration_functor_type::result_type restoration_result_type;

		context().m_file_output.normal( starttime(), m_mean_mz );

		if( m_indices.size() > 1 ) {
			double dimm_seeing = 0.0;
			double dimm_turbulence = 0.0;
			double dimm_seeing_err = 0.0;
			double dimm_turbulence_err = 0.0;

			if( m_dimm.size() > 1 ) {
				dimm::vars_type xvars = mean( m_dimm.xvars().begin(), m_dimm.xvars().end() );
				dimm::vars_type yvars = mean( m_dimm.yvars().begin(), m_dimm.yvars().end() );

				/* low vars correction */
				dimm::correct_low_vars( m_dimm.xvars().begin(), m_dimm.xvars().end(), xvars(0) );
				dimm::correct_low_vars( m_dimm.yvars().begin(), m_dimm.yvars().end(), yvars(0) );
				xvars = mean( m_dimm.xvars().begin(), m_dimm.xvars().end() );
				yvars = mean( m_dimm.yvars().begin(), m_dimm.yvars().end() );
				
				dimm::vars_type xerrs = error_of_mean( m_dimm.xvars().begin(), m_dimm.xvars().end(), xvars );
				dimm::vars_type yerrs = error_of_mean( m_dimm.yvars().begin(), m_dimm.yvars().end(), yvars );

				double Jlong = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Klongitudinal(), xvars(1) );
				double Jlong_err = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Klongitudinal(), xerrs(1) );
				double Jtrans = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Ktransversal(),  yvars(1) );
				double Jtrans_err = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Ktransversal(),  yerrs(1) );

				dimm_turbulence = 0.5*( Jlong + Jtrans ) / m_mean_mz;
				dimm_turbulence_err = std::sqrt( Jlong_err*Jlong_err + Jtrans_err*Jtrans_err ) / (Jlong + Jtrans);
				dimm_seeing = dimm::seeing( dimm_turbulence );
				dimm_seeing_err = 0.6*dimm_turbulence_err;
			}

			indices::flux_type flux_avg = mean( m_indices.flux().begin(), m_indices.flux().end() );
			indices::flux_type flux_err = error_of_mean( m_indices.flux().begin(), m_indices.flux().end(), flux_avg );
			indices::indices_type indices_avg = mean( m_indices.indices().begin(), m_indices.indices().end() );
			indices::indices_type indices_err = error_of_mean( m_indices.indices().begin(), m_indices.indices().end(), indices_avg );
			indices::indesis_type indesis_avg = mean( m_indices.indesis().begin(), m_indices.indesis().end() );
			indices::indesis_type indesis_err = error_of_mean( m_indices.indesis().begin(), m_indices.indesis().end(), indesis_avg );

			context().m_file_output.indices( meantime, indices_avg, indices_err, indesis_avg, indesis_err );
			context().m_file_output.fluxes( meantime, flux_avg, flux_err );

			integrals_functor_type ifun = context().m_integrals_factory( context().curobject().sclass(), context().curobject().weight(), context().curobject().short_weight() );

			integrals::storage integrals;
			for( indices::storage::const_iterator it = m_indices.begin(); it != m_indices.end(); ++it ) {
				const std::tuple<integrals::moments_type,integrals::moments_type,integrals::desi_type>& moments = ifun( it->indices, it->wind_indices, it->indesis, m_mean_mz );

				integrals.push_back( std::get<0>(moments), std::get<1>(moments), std::get<2>(moments) );
			}

			integrals::moments_type moments = mean( integrals.moments().begin(), integrals.moments().end() );
			integrals::moments_type moments_err = error_of_mean( integrals.moments().begin(), integrals.moments().end(), moments );
			integrals::moments_type wind_moments = mean( integrals.wind_moments().begin(), integrals.wind_moments().end() );
			integrals::moments_type wind_moments_err = error_of_mean( integrals.wind_moments().begin(), integrals.wind_moments().end(), wind_moments );
			integrals::desi_type desi = mean( integrals.desi().begin(), integrals.desi().end() );
			integrals::desi_type desi_err = error_of_mean( integrals.desi().begin(), integrals.desi().end(), desi );

			double moment_0 = moments(integrals::moment_0);
			double moment_2 = moments(integrals::moment_2);
			double moment_0_err = moments_err(integrals::moment_0)/moments(integrals::moment_0);
			double moment_2_err = moments_err(integrals::moment_2)/moments(integrals::moment_2);
			
			std::pair<double,double> free_seeing = integrals::seeing(moment_0,moment_0_err);
			std::pair<double,double> isoplanatic_angle = integrals::isoplanatic_angle(moments(integrals::moment_53),moments_err(integrals::moment_53)/moments(integrals::moment_53));
			std::pair<double,double> altitude_eff = integrals::effective_altitude(moment_0,moment_0_err,moments(integrals::moment_1),moments_err(integrals::moment_1)/moments(integrals::moment_1));
			std::pair<double,double> desi_time_constant = integrals::time_constant_desi(desi,desi_err/desi,m_exposition);
			std::pair<double,double> time_constant = integrals::time_constant(wind_moments(integrals::moment_0),wind_moments_err(integrals::moment_0)/wind_moments(integrals::moment_0),moment_0,moment_0_err,m_exposition);

			context().m_file_output.atmosphere( meantime,
				free_seeing.first, free_seeing.second,
				dimm_seeing, dimm_seeing_err,
				dimm_turbulence, dimm_turbulence_err,
				isoplanatic_angle.first, isoplanatic_angle.second,
				altitude_eff.first, altitude_eff.second,
				desi_time_constant.first, desi_time_constant.second,
				time_constant.first, time_constant.second,
				moment_0, moment_0_err,
				moment_2, moment_2_err );

			/* mixture */
			mixture::storage mixture_samples( m_indices, m_dimm, m_dimm.size() > 1 ? context().m_pproxy.dimm_scale() : 0 );

			vector< double > grid( context().m_va["profile-grid"].as< profile_grid >().vec );

			if( m_dimm.size() > 1 ) {
				grid.resize( grid.size() + 1 );
				subrange( grid, 1, grid.size() ) = subrange( grid, 0, grid.size() - 1 );
				grid(0) = 0;
			}
			
			/* covariate mixture */
			mixture::mixture_type mixture_avg = mean( mixture_samples.mixture().begin(), mixture_samples.mixture().end() );
			symmetric_matrix< double > mixture_covariance = covariance( mixture_samples.mixture().begin(), mixture_samples.mixture().end(), mixture_avg );
			/* prepare grid */

			restoration_functor_type rfun(
				grid,
				mixture_covariance,
				m_mean_mz,
				context().curobject().weight()
				);

			profile::storage profile;
			for( mixture::storage::const_iterator it = mixture_samples.begin(); it != mixture_samples.end(); ++it ) {
				profile.push_back( rfun( it->mixture ) );
			}

			profile.unstable_remove_if( profile::filter::chi2_gt_than( context().m_va["chi2-threshold"].as<double>() ) );

			if( profile.size() ) {
				double chi2_avg = mean( profile.chi2().begin(), profile.chi2().end() );
				vector< double > cn2_avg = mean( profile.cn2().begin(), profile.cn2().end() );
				vector< double > cn2_err = error_of_mean( profile.cn2().begin(), profile.cn2().end(), cn2_avg );

				context().m_file_output.profile( meantime, "X", rfun.seeing( cn2_avg ), chi2_avg, cn2_avg, cn2_err, grid );

				vector< double > residual_avg = mean( profile.residual().begin(), profile.residual().end() );
				residual_avg = element_div( rfun.indices_residual( residual_avg ), mixture_avg );

				boost::optional< std::pair<double, double> > residual_dimm;
				if( m_dimm.size() > 1 ) { residual_dimm = weight::unpack_dimm(residual_avg); }
				context().m_file_output.residual( meantime, "X", weight::as_symmetric(weight::unpack_mass(residual_avg)), residual_dimm );

			}
		}

		m_accumulation_begin = m_accumulation_end = boost::posix_time::ptime();
		m_mean_mz = 0;
		m_num = 0;
	}
};

double worksheet::object::airmass( const time_type& time ) const {
	yalpa::horizon<double> hc( yalpa::make_horizon_helper( m_coords, context().m_pproxy.site(), yalpa::sidereal_time(time) ) );
	double secz1 = 1.0/( std::cos(hc.coaltitude()) );
	double secz2 = secz1 - 1.0;
	return secz1 - secz2*( 0.0018167 + secz2*( 0.002875 + 0.0008083*secz2 ) );
}
const worksheet::object::nullexposure_weight_type& worksheet::object::weight() {
	return context().m_weight_factory.nullexposure_weight(
		m_spectral,
		context().m_pproxy.inner_diam(),
		context().m_pproxy.outer_diam(),
		context().m_pproxy.dimm_aperture_base(),
		context().m_pproxy.dimm_aperture_size() );
}
const worksheet::object::shortexposure_weight_type& worksheet::object::short_weight() {
	return context().m_weight_factory.shortexposure_weight(
		m_spectral,
		context().m_pproxy.inner_diam(),
		context().m_pproxy.outer_diam() );
}

void validate(boost::any& v, const std::vector<std::string>& values, worksheet::profile_grid* target_type, int) {
	namespace po = boost::program_options;
	using namespace boost::numeric;
	using namespace utils::algo;
	using namespace utils::ublas;

	typedef std::vector< std::string > split_vector_type;

	std::string s = po::validators::get_single_string(values);
	split_vector_type split_vec;

	try{
		boost::trim_if( s, boost::is_space() );
		boost::split( split_vec, s, boost::is_any_of(", "), boost::token_compress_on );
		if( split_vec.size() == 4 ){
			double start = boost::lexical_cast<double>( split_vec[0] );
			double stop  = boost::lexical_cast<double>( split_vec[1] );
			double step  = boost::lexical_cast<double>( split_vec[2] );
			double step2 = boost::lexical_cast<double>( split_vec[3] );
			if( step2 == 0.0 ){
				linear_generator gen( start, step );
				v = boost::any( worksheet::profile_grid(vector_generator< linear_generator > ( gen, gen.size( stop ) )) );
			} else {
				lg_generator gen( start, step2, step );
				v = boost::any( worksheet::profile_grid(vector_generator< lg_generator > ( gen, gen.size( stop ) )) );
			}
		} else {
			ublas::vector<double> vec( split_vec.size() );
			for( split_vector_type::size_type it = 0; it < split_vec.size(); ++it ) {
				vec(it) = boost::lexical_cast<double>( split_vec[it] );
			}
			v = boost::any( worksheet::profile_grid(vec) );
		}
	} catch( const std::exception& e ) {
	#if BOOST_VERSION < 104200
		throw po::validation_error("invalid value");
	#else
		throw po::validation_error( po::validation_error::invalid_option_value );
	#endif
	}
}

int worksheet::parse_program_options( int argc, char** argv, boost::program_options::variables_map& va ) {
	namespace po = boost::program_options;

	po::options_description desc("Allowed options");
	desc.add_options()
		( "help,h",    "produce help message" )
		( "version,v", "produce version message" )
		( "accum-time,a",    po::value<double>()->default_value(60.0), "force accumulation time" )
		( "background-threshold,b", po::value<double>()->default_value(4.0), "set background threshold" )
		( "chi2-threshold,x", po::value<double>()->default_value(25.0), "set chi2 threshold" )
		( "enable-simple-background", "do not have regard to the sun's altitude")
		( "flux-threshold",  po::value<double>()->default_value(2.0), "set flux threshold" )
		( "profile-grid,Z",  po::value<profile_grid>()->default_value( profile_grid( utils::ublas::vector_generator< utils::algo::lg_generator >( utils::algo::lg_generator( 0.5, 1.0, 2.0 ),6 ) ), "0.5,1,2,4,8,16" ), "set for profile grid generator" )
		( "mass-response,r", po::value<std::string>()->default_value("mass_maid.crv"), "file with mass detector response" )
		( "ccd-response,c",  po::value<std::string>()->default_value("ccd.crv"), "file with ccd response" )
		( "dump-weight,w",   "dump weight functions to the files")
		( "spectra-dir,s",   po::value<std::string>()->default_value("spectra"), "spectra-files directory" );

	try {
		po::store( po::parse_command_line( argc, argv, desc ), va );
		po::notify( va );
		if( va.count("version") ) {
			std::cerr << (boost::format("%1% %2%") % PACKAGE % VERSION) << std::endl;
			std::cerr << "Copyright (C) 2009-2010  Sternberg Astronomical Institute, MSU" << std::endl;
			std::cerr << "This is free software; see the source for copying conditions.  There is NO" << std::endl;
			std::cerr << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << std::endl;
			return 1;
		}
		if( va.count("help") ) {
			std::cerr << desc << std::endl;
			return 1;
		}
	} catch ( const po::error& e ) {
		std::cerr << e.what() << std::endl;
		std::cerr << desc << std::endl;
		return 1;
	}
	return 0;
}
atmos::background::model::abstract* worksheet::select_background_model_smart( const boost::posix_time::ptime& time ) const {
	using namespace std::placeholders;
	using namespace atmos::background::model;
	
	double alt = sun_altitude( time ) * 180 / M_PI; // degrees

	if( alt > -5.0 )
		return new daytime( time );
	else if( alt > -12.0 )
		return new twilight( time, std::bind(std::mem_fn(&worksheet::sun_altitude), this, _1) );
	// else
	return new night( time );
}
atmos::background::model::abstract* worksheet::select_background_model_simple( const boost::posix_time::ptime& time ) const {
	using namespace atmos::background::model;
	return new night( time );
}
double worksheet::sun_altitude( const boost::posix_time::ptime& time ) const {
	yalpa::ecliptic<double> sun = yalpa::sun_position( time );
	yalpa::horizon<double> sun_hz( make_horizon_helper( sun, m_pproxy.site(), yalpa::sidereal_time(time) ) );
	return 0.5 * M_PI - sun_hz.coaltitude();
}
worksheet::worksheet( const boost::program_options::variables_map& va ):
	m_va(va),
	m_pcontainer(),
	m_pproxy( m_pcontainer ),
	m_weight_factory( boost::filesystem::path( va["spectra-dir"].as<std::string>() ), boost::filesystem::path( va["mass-response"].as<std::string>() ), boost::filesystem::path( va["ccd-response"].as<std::string>() ) ),
	m_background_factory( std::bind( std::mem_fn( va.count("enable-simple-background") ? &worksheet::select_background_model_simple : &worksheet::select_background_model_smart ), this, std::placeholders::_1) ),
	m_file_input( std::cin, *this ),
	m_file_output( std::cout )
{
}

void worksheet::pre() {
	/* do nothing */
	m_file_output.comment((boost::format("%1% %2%") % PACKAGE % VERSION).str());
}
void worksheet::run() {
	bool ret = true;
	while( ret ) {
		try {
			ret &= m_file_input.shift();
		} catch ( const std::exception& e ) {
			std::cerr << m_file_input.line() << ": " << e.what() << std::endl;
		}
	}
}
void worksheet::post() {
	if( m_va.count("dump-weight") ) {
		for( weight_factory_type::nullexposure_factory_type::const_iterator it = weight_factory().nullexposure_begin(); it != weight_factory().nullexposure_end(); ++it ) {
			atmos::weight::dump_weight( ( boost::format("weight-%1%.wf") % it->first ).str(), it->second );
		}
		for( weight_factory_type::shortexposure_factory_type::const_iterator it = weight_factory().shortexposure_begin(); it != weight_factory().shortexposure_end(); ++it ) {
			atmos::weight::dump_weight( ( boost::format("weight-short-%1%.wf") % it->first ).str(), it->second );
		}
	}
}
worksheet::~worksheet() {

}

void worksheet::input::on_comment( const std::string& line ) {
	/* do nothing */
}
void worksheet::input::on_parameter( const time_type& time, const std::string& key, const std::string& value ) {
	context().m_pcontainer[key] = value;

	context().m_file_output.parameter( time, key, value );
}
void worksheet::input::on_object( const time_type& time, const object_type& obj ) {
	if( accumulator_type< dummy_accumulator >() ) {
		context().m_curobject = std::unique_ptr< context_type::object >( new context_type::object(obj.coordinates,obj.sclass,context()) );

		context().m_file_output.objectinfo( time, obj );
	} else {
		unshift();
		switch_accumulator<dummy_accumulator>( time );
	}
}
void worksheet::input::on_background( const time_type& time ) {
	if( accumulator_type< dummy_accumulator >() ) {
		switch_accumulator<background_accumulator>( time );
	} else {
		unshift();
		switch_accumulator<dummy_accumulator>( time );
	}
}
void worksheet::input::on_normal( const time_type& time ) {
	if( accumulator_type< dummy_accumulator >() ) {
		switch_accumulator<normal_accumulator>( time );
	} else {
		if( !accumulator_type< normal_accumulator >() || (( time - m_maccum->starttime() ).total_seconds() >= context().m_va["accum-time"].as<double>() ) ) {
			unshift();
			switch_accumulator<dummy_accumulator>( time );
		}
	}
}
void worksheet::input::on_flux( const time_type& time ) {
	switch_accumulator<dummy_accumulator>( time ); // dummy no throw
}
void worksheet::input::on_unknown_mode( const time_type& time ) {
	switch_accumulator<dummy_accumulator>( time ); // dummy no throw
}
void worksheet::input::on_measurements( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
	m_maccum->accumulate( time, mean, vars, rho1, rho2 );
}
void worksheet::input::on_dimm( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {
	m_maccum->dimm_accumulate( time, framenum, left, right, xvars, yvars );
}

}
