/*
    $Id: worksheet.h 323 2024-10-30 12:07:25Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _WORKSHEET_H
#define _WORKSHEET_H

#include <boost/program_options.hpp>

#include "background_factory.h" // for background_factory
#include "file_input.h"       // for file_input
#include "file_output.h"      // for file_output
#include "indices.h"
#include "integrals.h"
#include "local_context.h"
#include "params.h"           // for parameters_*
#include "profile.h"
#include "weight_factory.h"   // for weight::factory

namespace atmos {

using utils::local_context;

class worksheet {
	public:
		typedef local_context<worksheet> context_type;
		typedef atmos::parameters::container      parameters_container_type;
		typedef atmos::parameters::proxy          parameters_proxy_type;
		typedef atmos::weight::factory            weight_factory_type;
		typedef atmos::background::factory        background_factory_type;
		typedef atmos::integrals::functor_factory integrals_factory_type;
	public:
		static int parse_program_options( int argc, char** argv, boost::program_options::variables_map& va );
	public:
		struct profile_grid;
	private:
		class input:
			public atmos::input::file,
			public local_context< worksheet > {
		private:
			class measurement_accumulator: public local_context< worksheet > {
			private:
				time_type m_starttime;
			public:
				measurement_accumulator( const time_type& time, context_reference_type ctx ) throw() : local_context_type( ctx ), m_starttime(time) {};
				virtual ~measurement_accumulator() = 0;
				virtual void accumulate( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) = 0;
				virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) = 0;
				virtual void flush() = 0;

				template<class T> bool type() const {
					return dynamic_cast<const T*>(this);
				}

				inline const time_type& starttime() const { return m_starttime; }
			};

			class dummy_accumulator: public measurement_accumulator {
			public:
				dummy_accumulator( const time_type& time, context_reference_type ctx ) throw(): measurement_accumulator( time, ctx ) {};
				virtual ~dummy_accumulator() {};
				virtual void accumulate(  const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {}
				virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {}
				virtual void flush() {}
			};

			class background_accumulator;
			class normal_accumulator;
		private:
			std::unique_ptr<measurement_accumulator> m_maccum;

			template<class T> void switch_accumulator( const time_type& time ) {
				std::unique_ptr< measurement_accumulator > new_accumulator( new T( time,context() ) );
				std::swap( m_maccum, new_accumulator );
				new_accumulator->flush();
			}

			template<class T> bool accumulator_type() const {
				return m_maccum->type<T>();
			}
		public:
			void on_comment( const std::string& line );
			void on_parameter( const time_type& time, const std::string& key, const std::string& value );
			void on_object( const time_type& time, const object_type& obj );
			void on_background( const time_type& time );
			void on_normal( const time_type& time );
			void on_flux( const time_type& time );
			void on_unknown_mode( const time_type& time );
			void on_measurements( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 );
			void on_dimm( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars );
		public:
			input( std::istream& stm, context_reference_type ctx ): atmos::input::file( stm ), local_context_type( ctx ), m_maccum( new dummy_accumulator( time_type() ,ctx) ) {}
		};
		
		class object;
		
		friend class input;
		friend class object;
	private:
		const boost::program_options::variables_map& m_va;
		
		parameters_container_type m_pcontainer;
		parameters_proxy_type     m_pproxy;
		weight_factory_type       m_weight_factory;
		background_factory_type   m_background_factory;
		integrals_factory_type    m_integrals_factory;
		
		input        m_file_input;
		output::file m_file_output;

		std::unique_ptr<object> m_curobject;
	private:
		background::model::abstract* select_background_model_smart( const background::time_type& time ) const;
		background::model::abstract* select_background_model_simple( const background::time_type& time ) const;
	public:
		worksheet( const boost::program_options::variables_map& va );
		~worksheet( );

		inline const parameters_proxy_type&   parameters()         const { return m_pproxy;   }
		inline const weight_factory_type&     weight_factory()     const { return m_weight_factory; }
		inline const background_factory_type& background_factory() const { return m_background_factory; }
		inline const integrals_factory_type&  integrals_factory()  const { return m_integrals_factory; }

		double sun_altitude( const background::time_type& time ) const;

		inline object& curobject() {
			if( ! m_curobject.get() )
				throw std::logic_error("There is not any object");
			return *m_curobject;
		}

		void pre();
		void run();
		void post();
};

}

#endif // _WORKSHEET_H
