/*
    $Id: io.h 16 2011-10-25 07:38:38Z matwey $
    Copyright (C) 2010  Matwey V. Kornilov <matwey.kornilov@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _YALPA_IO_H
#define _YALPA_IO_H

#include <cmath>
#include <iosfwd>
#include <iostream>

#include <boost/math/special_functions/sign.hpp>

namespace yalpa {
	
	template<class T> struct angle {
		typedef T                 value_type;
		typedef value_type&       reference_type;
		typedef const value_type& const_reference_type;
		
		angle( reference_type val ): angle_(val) {
		}
		operator T () const {
			return angle_;
		}
		const_reference_type operator() () const {
			return angle_;
		}
		reference_type operator() () {
			return angle_;
		}
		private:
			reference_type angle_;
	};

	template<class T> angle<T> make_angle( T& val ){
		return angle<T>(val);
	}
	template<class T> angle<const T> make_angle( const T& val ){
		return angle<const T>(val);
	}
	
};

namespace std {

	std::ostream& degrees( std::ostream& stm );
	std::istream& degrees( std::istream& stm );
	std::ostream& hours( std::ostream& stm );
	std::istream& hours( std::istream& stm );
	std::ostream& radians( std::ostream& stm );
	std::istream& radians( std::istream& stm );

	std::ostream& decimal( std::ostream& stm );
	std::istream& decimal( std::istream& stm );
	std::ostream& sexagesimal( std::ostream& stm );
	std::istream& sexagesimal( std::istream& stm );

	bool is_degrees( std::ostream& stm );
	bool is_degrees( std::istream& stm );
	bool is_hours( std::ostream& stm );
	bool is_hours( std::istream& stm );
	bool is_sexagesimal( std::ostream& stm );
	bool is_sexagesimal( std::istream& stm );
	
	template<class T> std::ostream& operator<<(std::ostream& stm, const yalpa::angle<T>& val ) {
		std::ostream::sentry se(stm);
		if( se ){
			if( is_degrees(stm) ){
				T degree = val()/M_PI*180;
				if( is_sexagesimal(stm) ) {
					if( val() < 0 ) { stm << "-"; }
					stm << int(std::abs(degree)) << stm.fill() << int(std::abs(60*std::fmod(degree,1.0))) << stm.fill() << std::abs((3600*std::fmod(degree,1.0/60.0)));
				} else {
					stm << degree;
				}
			}else if( is_hours(stm) ){
				T hour = val()/M_PI*12;
				if( is_sexagesimal(stm) ) {
					if( val() < 0 ) { stm << "-"; }
					stm << int(std::abs(hour)) << stm.fill() << int(std::abs(60*std::fmod(hour,1.0))) << stm.fill() << std::abs((3600*std::fmod(hour,1.0/60.0)));
				} else {
					stm << hour;
				}
			}else{
				stm << val(); // in rad
			}
		}
		return stm;
	}
	template<class T> std::istream& operator>>( std::istream& stm, yalpa::angle<T> val ) {
		using boost::math::copysign;
		
		std::istream::sentry se(stm);
		if( se ){
			if( is_degrees(stm) ){
				T degree;
				if( is_sexagesimal(stm) ) {
					T deg,min,sec;
					stm >> deg >> min >> sec;
					val() = copysign( ( std::abs(deg) + min/60.0 + sec/3600.0 ) / 180.0 * M_PI, deg );
				} else {
					stm >> degree;
					val() = degree / 180.0 * M_PI;
				}
			}else if( is_hours(stm) ){
				T hour;
				if( is_sexagesimal(stm) ) {
					T hour,min,sec;
					stm >> hour >> min >> sec;
					val() = copysign( ( std::abs(hour) + min/60.0 + sec/3600.0 ) / 12.0 * M_PI, hour );
				} else {
					stm >> hour;
					val() = hour / 12.0 * M_PI;
				}
			}else{
				stm >> val(); // in rad
			}
		}
		return stm;
	}
	
};

#endif // _YALPA_IO_H
