/*
    $Id: point.h 36 2014-03-12 15:50:43Z matwey $
    Copyright (C) 2010  Matwey V. Kornilov <matwey.kornilov@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _POINT_H
#define _POINT_H

#include <cmath>

#include <boost/numeric/ublas/vector.hpp>

namespace yalpa {

	template<class T> struct point_expression {
		typedef T                      expression_type;
		typedef expression_type&       reference_type;
		typedef const expression_type& const_reference_type;
	};
	
	template<class T> class point {
		public:
			typedef T                 value_type;
			typedef value_type&       reference_type;
			typedef const value_type& const_reference_type;
			typedef point<T>          self_type;
			typedef boost::numeric::ublas::vector< value_type, boost::numeric::ublas::bounded_array< value_type, 3 > > storage_type;
		private:
			point( const storage_type& vec ): vec_(vec) {
			}
		public:
			point( const_reference_type x, const_reference_type y, const_reference_type z ): vec_(3) {
				vec_(0) = x;
				vec_(1) = y;
				vec_(2) = z;
			}
			point( const point& p ): vec_( p.vec_ ) {
			}
			point(): vec_(3) {
				vec_(0) = 0;
				vec_(1) = 0;
				vec_(2) = 0;
			}
			template< class E > point( const point_expression<E>& pe ): vec_(3) {
				vec_(0) = reinterpret_cast< typename point_expression<E>::const_reference_type >(pe).x();
				vec_(1) = reinterpret_cast< typename point_expression<E>::const_reference_type >(pe).y();
				vec_(2) = reinterpret_cast< typename point_expression<E>::const_reference_type >(pe).z();
			}
			~point() {
			}
	
			point& operator= ( const point& p ) {
				vec_ = p.vec_;
				return *this;
			}
			template< class E > point& operator=( const point_expression<E>& pe ) {
				self_type npoint( pe );
				std::swap( *this, npoint );
				return *this;
			}
	
			inline const_reference_type x() const { return vec_(0); }
			inline const_reference_type y() const { return vec_(1); }
			inline const_reference_type z() const { return vec_(2); }

			inline value_type inner_prod( const point& p ) const {
				return boost::numeric::ublas::inner_prod(vec_, p.vec_);
			}
		private:
			storage_type vec_;
	};

	template< class T > typename point<T>::value_type inner_prod( const point<T>& x, const point<T>& y ) {
		return x.inner_prod(y);
	}
	
	template< class P > class rotate_x:
		public point_expression< rotate_x< P > > {
			
		public:
			typedef P                 point_type;
			typedef typename point_type::value_type value_type;
			typedef value_type&       reference_type;
			typedef const value_type& const_reference_type;
			typedef rotate_x<P>       self_type;
		private:
		public:
			rotate_x( const point_type& point, const_reference_type angle ): point_(point), angle_(angle) {
			};
			rotate_x& operator= ( const point_type& point ) { point_ = point; }
	
			inline value_type x() const { return point_.x(); }
			inline value_type y() const { return  std::cos(angle_) * point_.y() + std::sin(angle_) * point_.z(); }
			inline value_type z() const { return -std::sin(angle_) * point_.y() + std::cos(angle_) * point_.z(); }
		private:
			const point_type& point_;
			value_type angle_;
	};

	template< class P > class rotate_y:
		public point_expression< rotate_y< P > > {
			
		public:
			typedef P                 point_type;
			typedef typename point_type::value_type value_type;
			typedef value_type&       reference_type;
			typedef const value_type& const_reference_type;
			typedef rotate_y<P>       self_type;
		private:
		public:
			rotate_y( const point_type& point, const_reference_type angle ): point_(point), angle_(angle) {
			};
			rotate_y& operator= ( const point_type& point ) { point_ = point; }
	
			inline value_type x() const { return std::cos(angle_) * point_.x() - std::sin(angle_) * point_.z(); }
			inline value_type y() const { return point_.y(); }
			inline value_type z() const { return std::sin(angle_) * point_.x() + std::cos(angle_) * point_.z(); }
		private:
			const point_type& point_;
			value_type angle_;
	};

	template< class P > class rotate_z:
		public point_expression< rotate_z< P > > {
			
		public:
			typedef P                 point_type;
			typedef typename point_type::value_type value_type;
			typedef value_type&       reference_type;
			typedef const value_type& const_reference_type;
			typedef rotate_z<P>       self_type;
		private:
		public:
			rotate_z( const point_type& point, const_reference_type angle ): point_(point), angle_(angle) {
			};
			rotate_z& operator= ( const point_type& point ) { point_ = point; }
	
			inline value_type x() const { return  std::cos(angle_) * point_.x() + std::sin(angle_) * point_.y(); }
			inline value_type y() const { return -std::sin(angle_) * point_.x() + std::cos(angle_) * point_.y(); }
			inline value_type z() const { return point_.z(); }
		private:
			const point_type& point_;
			value_type angle_;
	};

	template< class P > rotate_x< P > make_rotate_x( const P& point, const typename P::value_type& angle ) {
		return rotate_x< P >( point, angle );
	}
	template< class P > rotate_y< P > make_rotate_y( const P& point, const typename P::value_type& angle ) {
		return rotate_y< P >( point, angle );
	}
	template< class P > rotate_z< P > make_rotate_z( const P& point, const typename P::value_type& angle ) {
		return rotate_z< P >( point, angle );
	}
};

#endif // _POINT_H
