/*
    $Id: spheric.h 36 2014-03-12 15:50:43Z matwey $
    Copyright (C) 2010  Matwey V. Kornilov <matwey.kornilov@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _SPHERIC_H
#define _SPHERIC_H

#include "point.h"

#include <cmath>

namespace yalpa {
	
	template<class T> class spheric {
		public:
			typedef T                   value_type;
			typedef value_type&         reference_type;
			typedef const value_type&   const_reference_type;
			typedef spheric<T>          self_type;
			typedef point< value_type > point_type;
		private:
			point_type construct_point( const_reference_type angle1, const_reference_type angle2 ) {
				return point_type( std::sin(angle2)*std::cos(angle1) , std::sin(angle2)*std::sin(angle1) , std::cos(angle2) );
			}
		public:
			spheric( const_reference_type angle1, const_reference_type angle2 ):
				point_( construct_point( angle1, angle2 ) ) {
			}
			spheric( const spheric& s ): point_( s.point_ ) {
			}
			self_type& operator= ( const spheric& s ) {
				point_ = s.point_;
				return *this;
			}
			spheric( const point_type& p ): point_( p ) {
			}
			self_type& operator= ( const point_type& p ) {
				point_ = p;
			}

			inline value_type angle1() const {
				return std::atan2(point_.y(),point_.x());
			}
			inline value_type angle2() const {
				return std::acos(point_.z());
			}
			inline void angle1( const_reference_type angle1 ) {
				point_ = construct_point( angle1, angle2() );
			}
			inline void angle2( const_reference_type angle2 ) {
				point_ = construct_point( angle1(), angle2 );
			}

			inline const point_type& getpoint() const { return point_; }
		private:
			point_type point_;
	};

	/**
		@brief return angular distance between two points
		@param x - first point
		@param y - second point
		@return angular distance between x and y
	*/
	template<class T> typename spheric<T>::value_type distance( const spheric<T>& x, const spheric<T>& y ) {
		return std::acos(inner_prod(x.getpoint(),y.getpoint()));
	}

	template<class T> class geographic;
	template<class T> class ecliptic;
	template<class T> class equatorial;
	template<class T> class horizon;
	template<class T> struct horizon_helper;
	
	template<class T> class geographic:
		public spheric<T> {
		public:
			typedef typename spheric<T>::value_type           value_type;
			typedef typename spheric<T>::reference_type       reference_type;
			typedef typename spheric<T>::const_reference_type const_reference_type;
			typedef geographic<T>                             self_type;
			typedef typename spheric<T>::point_type           point_type;
		public:
			geographic( const_reference_type longitude = 0, const_reference_type latitude = 0 ):
				spheric<T>( ( longitude >= 0 ? longitude : longitude + 2 * M_PI ), 0.5 * M_PI - latitude ) {
			}

			inline value_type longitude() const {
				return this->angle1();
			}
			inline value_type latitude() const {
				return 0.5 * M_PI - this->angle2();
			}
			inline void longitude( const_reference_type longitude ) {
				this->angle1( ( longitude >= 0 ? longitude : longitude + 2 * M_PI ) );
			}
			inline void latitude( const_reference_type latitude ) {
				this->angle2( 0.5 * M_PI - latitude );
			}
	};
	
	template<class T> class equatorial:
		public spheric<T> {
		public:
			typedef typename spheric<T>::value_type           value_type;
			typedef typename spheric<T>::reference_type       reference_type;
			typedef typename spheric<T>::const_reference_type const_reference_type;
			typedef equatorial<T>                             self_type;
			typedef typename spheric<T>::point_type           point_type;
		public:
			equatorial( const_reference_type ra = 0, const_reference_type dec = 0 ):
				spheric<T>( ra, 0.5 * M_PI - dec ) {
			}

			inline value_type ra() const {
				value_type ra = this->angle1();
				return ( ra < 0 ? ra + 2*M_PI : ra );
			}
			inline value_type dec() const {
				return 0.5 * M_PI - this->angle2();
			}
			inline void ra( const_reference_type ra ) {
				this->angle1( ra );
			}
			inline void dec( const_reference_type dec ) {
				this->angle2( 0.5 * M_PI - dec );
			}
			/**
				@brief return hour angle of the point
				@param sidtime - siderial time of the moment
				@return hour angle in rad at sidtime
			*/
			inline value_type ha(const value_type& sidtime) const {
				value_type ha = std::fmod(sidtime - ra(), 2*M_PI);
				return ( ha < 0 ? ha + 2*M_PI : ha );
			}
			/**
				@return hour angle in rad of the set at location (rise = 2 Pi - set)
			*/
			inline value_type ha(const geographic<value_type>& location) const {
				const value_type& dec_ = dec();
				const value_type& lat = location.latitude();
				if( dec_ + lat > 0.5*M_PI ) { /* non-setting */
					return std::numeric_limits<value_type>::infinity();
				} else if( dec_ - lat < -0.5*M_PI ) { /* non-rising */
					return -std::numeric_limits<value_type>::infinity();
				}
				value_type num  = std::sqrt(std::cos(lat - dec_));
				value_type dnum = std::sqrt(std::cos(lat + dec_));
				return std::atan2(num,dnum)*2 /* [0 to Pi] - set */;
			}
			
			equatorial( const ecliptic< value_type >& ecl );
			self_type& operator= ( const ecliptic< value_type >& ecl );
			equatorial( const horizon_helper< horizon< value_type >  >& hor );
			self_type& operator= (const horizon_helper< horizon< value_type >  >& hor );
	};

	template<class T> class ecliptic:
		public spheric<T> {
		public:
			typedef typename spheric<T>::value_type           value_type;
			typedef typename spheric<T>::reference_type       reference_type;
			typedef typename spheric<T>::const_reference_type const_reference_type;
			typedef ecliptic<T>                               self_type;
			typedef typename spheric<T>::point_type           point_type;
		public:
			ecliptic( const_reference_type longitude = 0, const_reference_type latitude = 0):
				spheric<T>( longitude, 0.5 * M_PI - latitude ) {
			}

			inline value_type longitude() const {
				double longitude = this->angle1();
				return ( longitude < 0 ? longitude + 2*M_PI : longitude );
			}
			inline value_type latitude() const {
				return 0.5 * M_PI - this->angle2();
			}
			inline void longitude( const_reference_type longitude ) {
				this->angle1( longitude );
			}
			inline void latitude( const_reference_type latitude ) {
				this->angle2( 0.5 * M_PI - latitude );
			}

			ecliptic( const equatorial< value_type >& eqt );
			self_type& operator= ( const equatorial< value_type >& eqt );
			ecliptic( const horizon_helper< horizon< value_type > >& hor );
			self_type& operator= (const horizon_helper< horizon< value_type > >& hor );
	};

	template<class T> class horizon:
		public spheric<T> {
		public:
			typedef typename spheric<T>::value_type           value_type;
			typedef typename spheric<T>::reference_type       reference_type;
			typedef typename spheric<T>::const_reference_type const_reference_type;
			typedef horizon<T>                                self_type;
			typedef typename spheric<T>::point_type           point_type;
		public:
			horizon( const_reference_type azimuth = 0, const_reference_type coaltitude = 0 ):
				spheric<T>( -azimuth, coaltitude ) {
			}

			inline value_type azimuth() const {
				return M_PI - this->angle1();
			}
			inline value_type coaltitude() const {
				return this->angle2();
			}
			inline void azimuth( const_reference_type azimuth ) {
				this->angle1( -azimuth );
			}
			inline void coaltitude( const_reference_type coaltitude ) {
				this->angle2( coaltitude );
			}

			horizon( const horizon_helper< equatorial< value_type > >& eqt );
			self_type& operator= ( const horizon_helper< equatorial< value_type > >& eqt );
			horizon( const horizon_helper< ecliptic< value_type > >& ecl );
			self_type& operator= ( const horizon_helper< ecliptic< value_type > >& ecl );
	};

	template<class C> struct horizon_helper {
		typedef C             system_type;
		typedef typename system_type::value_type value_type;
		typedef geographic<value_type> geographic_type;
			
		horizon_helper( const system_type& system, const geographic_type& geographic, value_type sidtime ):
			system_(system), geographic_(geographic), sidtime_(sidtime) {
		}
			
		const system_type&     system_;
		const geographic_type& geographic_;
		value_type             sidtime_;
	};
	template<class C> horizon_helper< C > make_horizon_helper( const C& system, const typename horizon_helper<C>::geographic_type& geographic, typename horizon_helper<C>::value_type sidtime ) {
		return horizon_helper< C > ( system, geographic, sidtime );
	}

	template<class T> equatorial<T>::equatorial( const ecliptic< value_type >& ecl ):
		spheric<T>( make_rotate_x( ecl.getpoint(), -(23.439281*M_PI/180) ) ) {
	}
	template<class T> typename equatorial<T>::self_type& equatorial<T>::operator= ( const ecliptic< value_type >& ecl ) {
		self_type tmp( ecl );
		std::swap( *this, tmp );
		return *this;
	}
	template<class T> equatorial<T>::equatorial( const horizon_helper< horizon< value_type > >& hor ):
		spheric<T>( make_rotate_z( make_rotate_x( make_rotate_z( hor.system_.getpoint(), -0.5 * M_PI ), hor.geographic_.angle2() ), 0.5 * M_PI - (hor.sidtime_+hor.geographic_.angle1()) ) ) {
	}
	template<class T> typename equatorial<T>::self_type& equatorial<T>::operator= ( const horizon_helper< horizon< value_type > >& hor ) {
		self_type tmp( hor );
		std::swap( *this, tmp );
		return *this;
	}

	template<class T> ecliptic<T>::ecliptic( const equatorial< value_type >& eqt ):
		spheric<T>( make_rotate_x( eqt.getpoint(), (23.439281*M_PI/180) ) ) {
	}
	template<class T> typename ecliptic<T>::self_type& ecliptic<T>::operator= ( const equatorial< value_type >& eqt ) {
		self_type tmp( eqt );
		std::swap( *this, tmp );
		return *this;
	}
	template<class T> ecliptic<T>::ecliptic( const horizon_helper< horizon< value_type > >& hor ):
		spheric<T>( make_rotate_x( make_rotate_z( make_rotate_x( make_rotate_z( hor.system_.getpoint(), -0.5 * M_PI ), hor.geographic_.angle2() ), 0.5 * M_PI - (hor.sidtime_+hor.geographic_.angle1()) ), (23.439281*M_PI/180) ) ) {
	}
	template<class T> typename ecliptic<T>::self_type& ecliptic<T>::operator= ( const horizon_helper< horizon< value_type > >& hor ) {
		self_type tmp( hor );
		std::swap( *this, tmp );
		return *this;
	}

	template<class T> horizon<T>::horizon( const horizon_helper< equatorial< value_type > >& eqt ):
		spheric<T>( make_rotate_z( make_rotate_x( make_rotate_z( eqt.system_.getpoint(), -0.5 * M_PI + (eqt.sidtime_+eqt.geographic_.angle1()) ), -eqt.geographic_.angle2() ), 0.5 * M_PI ) ) {
	}
	template<class T> typename horizon<T>::self_type& horizon<T>::operator= ( const horizon_helper< equatorial< value_type > >& eqt ) {
		self_type tmp( eqt );
		std::swap( *this, tmp );
		return *this;
	}
	template<class T> horizon<T>::horizon( const horizon_helper< ecliptic< value_type > >& ecl ):
		spheric<T>( make_rotate_z( make_rotate_x( make_rotate_z( make_rotate_x( ecl.system_.getpoint(), -(23.439281*M_PI/180) ), -0.5 * M_PI + (ecl.sidtime_+ecl.geographic_.angle1()) ), -ecl.geographic_.angle2() ), 0.5 * M_PI ) ) {
	}
	template<class T> typename horizon<T>::self_type& horizon<T>::operator= ( const horizon_helper< ecliptic< value_type > >& ecl ) {
		self_type tmp( ecl );
		std::swap( *this, tmp );
		return *this;
	}
};

#endif // _SPHERIC_H
