/*
    $Id: sidtime.cpp 44 2024-10-30 11:22:34Z matwey $
    Copyright (C) 2010  Matwey V. Kornilov <matwey.kornilov@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <yalpa/sidtime.h>

#include <cmath>

namespace {

template<class T> T inline mod(const T& x, const T& y) {
	if(y == T(0)) return x;
	return x - y * std::floor(x/y);
}

};

double yalpa::julian( const std::time_t& epoch ) {
	return 2440587.5 + ( epoch / 86400.0 );
}
/**
	@brief convert julian date to seconds since Jan 1, 1970
*/
std::time_t yalpa::epoch( double julian ) {
	return std::round((julian - 2440587.5) * 86400);
}
double yalpa::sidereal_time( const std::time_t& epoch ) {
	return sidereal_time( julian( epoch ) );
}
double yalpa::sidereal_time( double julian ) {
	double jd0 = std::floor(julian-0.5)+0.5;
	double t = (jd0 - 2451545.0) / 36525.0;
	double sid = (6*3600.0 + 41*60.0 + 50.54841) + 8640184.812866 * t + 0.093104 * t*t - 6.2e-6 * t*t*t + 1.002737909350795 * 86400.0 * (julian-jd0);
	return std::fmod( sid, 86400.0 ) * ( 2 * M_PI / 86400.0 );
}
/**
	@breif convert from sidereal time to julian date
	@param sid - sidereal time (in rad)
	@param julian0 - julian day
	@return julian date at sid moment
*/
double yalpa::from_sidereal( double sid, double julian0 ) {
	double jd0 = std::floor(julian0-0.5)+0.5;
	double t = (jd0 - 2451545.0) / 36525.0;
	double gmst0 = ((6*3600.0 + 41*60.0 + 50.54841) + 8640184.812866 * t + 0.093104 * t*t - 6.2e-6 * t*t*t)/86400;
	double delta_gmst = mod(sid/(2*M_PI) - gmst0,1.0);
	return jd0 + (delta_gmst - 0.002730433670915674 * delta_gmst);
}
double yalpa::julian( const boost::posix_time::ptime& time ) {
	return 2440587.5 + (time -  boost::posix_time::ptime( boost::gregorian::date( 1970,1,1 ) )).total_seconds() / 86400.0;
}
double yalpa::sidereal_time( const boost::posix_time::ptime& time ) {
	return sidereal_time( julian( time ) );
}
