/*
    $Id: sun.cpp 32 2013-08-04 15:27:24Z matwey $
    Copyright (C) 2010  Matwey V. Kornilov <matwey.kornilov@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <yalpa/sidtime.h>
#include <yalpa/sun.h>

#include <yalpa/io.h>

#include <cmath>

double yalpa::sun_longitude( const std::time_t& epoch ) {
	return sun_longitude( julian( epoch ) );
}

double yalpa::sun_longitude( double julian ) {
	/* from Jean Meeus */
	double t = (julian - 2451545.0) / 36525.0; // Julian centures
	double meanLongitude = M_PI / 180.0 * ( 280.46645 + 36000.76983 * t + 0.0003032 * t*t );
	double meanAnomaly   = M_PI / 180.0 * ( 357.52910 + 35999.05030 * t - 0.0001559 * t*t - 0.00000048 * t*t*t );
	double center        = M_PI / 180.0 * ( (1.914600 - 0.004817 * t - 0.000014 * t*t) * sin(meanAnomaly) + (0.019933 - 0.000101 * t) * sin(2*meanAnomaly) + 0.000290 * sin(3*meanAnomaly) );
	/* it is geometrical longitude. we don't corrent for nutation and the aberration */
	double longitude = std::fmod( meanLongitude + center, 2*M_PI );
	return ( longitude < 0 ? longitude + 2*M_PI : longitude );
}
yalpa::ecliptic<double> yalpa::sun_position( double julian ) {
	return ecliptic<double>( sun_longitude(julian), 0.0 );
}
yalpa::ecliptic<double> yalpa::sun_position( const std::time_t& epoch ) {
	return sun_position(julian(epoch));
}
double yalpa::sun_longitude( const boost::posix_time::ptime& time ) {
	return sun_longitude( julian( time ) );
}
yalpa::ecliptic<double> yalpa::sun_position( const boost::posix_time::ptime& time ) {
	return sun_position(julian(time));
}

namespace detail {

using namespace yalpa;

/**
	@brief return moment of the Sun set or ris
	@param jday - julian day to calculate on
	@param location - location of the observer
	@return julian moment of the set or rise

	eqt_sun.ha() returns hour angle = 2arctan(sqrt(...))
	(which is between 0 and Pi, and corresponds to the set)
	The second solution to the equation is the 2Pi - 2arctan(sqrt(...))
	(which is between Pi and 2Pi, and corresponds to the rise)
	The sign is just negated because 2Pi floored at from_sidereal()
*/
template<int sign> inline double sunset_time( double jday, const geographic<double>& location ) {
	double j = jday, j0 = jday;
	equatorial<double> eqt_sun(.0,.0);
	do {
		j0 = j;
		eqt_sun = sun_position(j);
		double st = sign*eqt_sun.ha(location) + eqt_sun.ra() - location.longitude();
		j = yalpa::from_sidereal(st, j);
	} while( std::abs(j0-j) > std::abs(j) * std::numeric_limits<double>::epsilon() );
	return j;	
}

};

double yalpa::sunset_time( double jday, const geographic<double>& location ) {
	return detail::sunset_time<1>(jday, location);
}
double yalpa::sunrise_time( double jday, const geographic<double>& location ) {
	return detail::sunset_time<-1>(jday, location);
}
std::time_t yalpa::sunset_time( const std::time_t& day, const geographic<double>& location ) {
	return yalpa::epoch(yalpa::sunset_time(yalpa::julian(day),location));
}
std::time_t yalpa::sunrise_time( const std::time_t& day, const geographic<double>& location ) {
	return yalpa::epoch(yalpa::sunrise_time(yalpa::julian(day),location));
}
